from .. import chemistry
from . import SELF, Details, Equation

PHYSICAL_CHEMISTRY: Details = {
    chemistry.KA: Equation(
        r"K_a = \frac{a_{\mathrm{H}^+} a_{\mathrm{A}^-}}{a_{\mathrm{HA}}}",
        {
            "K_a": SELF,
            r"a_{\mathrm{H}^+}": chemistry.ACTIVITY_HPLUS,
            r"a_{\mathrm{A}^-}": chemistry.ACTIVITY_AMINUS,
            r"a_{\mathrm{HA}}": chemistry.ACTIVITY_HA,
        },
        assumptions={"aqueous solution"},
    ),
}



"""United States customary and British Imperial units.

See: [isqx._citations.SP811][], [isqx._citations.H44][],
[isqx._citations.WMA1985][].
"""

from decimal import Decimal
from fractions import Fraction

from ._core import PI, LazyProduct, Translated
from ._iso80000 import (
    CONST_DENSITY_H2O,
    CONST_DENSITY_HG,
    CONST_STANDARD_GRAVITY,
    HOUR,
    KG,
    KGF,
    M_PERS2,
    MILLI,
    MIN,
    PA,
    G,
    J,
    K,
    L,
    M,
    S,
    W,
)

#
# temperature [H44 C-23]
#

R = (Fraction(5, 9) * K).alias("rankine")
"""Thermodynamic Temperature (Rankine). Absolute scale."""
FAHRENHEIT = Translated(R, Decimal("-459.67"), "fahrenheit")
"""Thermodynamic Temperature (Fahrenheit). Absolute, translated scale.
Cannot be composed with other units."""

#
# length [H44 C-24]
#

FT = (Decimal("0.3048") * M).alias("foot")
"""International foot."""
FT_US_SURVEY = (Fraction(1200, 3937) * M).alias("foot_us_survey")
"""U.S. survey foot, deprecated since Dec 31, 2022."""  # H44 C-10
IN = (Fraction(1, 12) * FT).alias("inch")
"""International inch."""
YD = (3 * FT).alias("yard")
"""International yard."""  # H44 C-25
MI = (5280 * FT).alias("mile")
"""International mile, also known as statute mile."""  # H44 C-25
NMI = (1852 * M).alias("nautical_mile")
"""International nautical mile."""  # H44 C-25
MIL = (Fraction(1, 1000) * IN).alias("mil")
"""Thousandth of an inch, a unit of thickness (also known as thou)."""
HAND = (4 * IN).alias("hand")
"""Unit of length, for measuring the height of horses."""
POINT = (Fraction(1, 72) * IN).alias("point")
"""Typographical point (desktop publishing)."""  # H44 C-25


# gunter's or surveyors chain units
ROD = (Fraction(33, 2) * FT).alias("rod")
"""Rod, also known as pole or perch."""  # H44 C-5, C-25
FURLONG = (40 * ROD).alias("furlong")
"""Furlong."""
FATHOM = (6 * FT).alias("fathom")
"""Fathom, a unit of length used for water depth."""
CABLE = (120 * FATHOM).alias("cable")
"""Cable length."""
LEAGUE = (3 * MI).alias("league")
"""League."""
LINK = (Fraction(66, 100) * FT).alias("link")
"""Gunter's link."""
CHAIN = (100 * LINK).alias("chain")
"""Gunter's chain."""

#
# area
#

SQ_YD = YD**2
ROOD = (1210 * SQ_YD).alias("rood")
"""Rood, an imperial unit of area."""  # WMA1985 1VI
SQ_FT = FT**2
ACRE = (43560 * SQ_FT).alias("acre")
"""International acre."""  # H44 C-25
ACRE_SURVEY = (43560 * (FT_US_SURVEY**2)).alias("acre_survey")
"""U.S. survey acre."""  # H44 C-14
SQ_BUILDING = (100 * SQ_FT).alias("square_building")
"""Square building, used in the U.S. construction industry."""
SQ_MI = MI**2
SQ_MIL = MIL**2
CIRCULAR_MIL = (LazyProduct((PI, (4, -1))) * SQ_MIL).alias("circular_mil")
"""Circular mil, a unit of area for wire cross-sections."""  # SP811 B.8

#
# volume [H44 C-6]
#

# liquid, US
CU_IN = IN**3
GAL = (231 * CU_IN).alias("gallon")
"""U.S. liquid gallon."""
QUART = (Fraction(1, 4) * GAL).alias("quart")
"""U.S. liquid quart."""
PINT = (Fraction(1, 2) * QUART).alias("pint")
"""U.S. liquid pint."""
CUP = (Fraction(1, 2) * PINT).alias("cup_measuring")
"""U.S. customary cup."""  # H44 C-27
GILL = (Fraction(1, 4) * PINT).alias("gill")
"""U.S. liquid gill."""
FL_OZ = (Fraction(1, 128) * GAL).alias("ounce_fluid")
"""U.S. fluid ounce."""
TABLESPOON = (Fraction(1, 2) * FL_OZ).alias("tablespoon_measuring")
"""U.S. customary tablespoon."""  # H44 C-28
TEASPOON = (Fraction(1, 3) * TABLESPOON).alias("teaspoon_measuring")
"""U.S. customary teaspoon."""  # H44 C-28
# consult federal and state laws for the appropriate barrel definition [H44 C-26]
BBL_FEDERAL_FERMENTED_LIQUOR = (31 * GAL).alias("bbl_federal_fermented_liquor")
BBL_STATE_LIQUID = ((31 + Fraction(1, 2)) * GAL).alias("bbl_state_liquid")
BBL_FEDERAL_CISTERN = (36 * GAL).alias("bbl_federal_cistern")
BBL_PROOF_SPIRIT = (50 * GAL).alias("bbl_proof_spirit")
BBL_OIL = (42 * GAL).alias("bbl_oil", allow_prefix=True)
"""U.S. standard barrel for crude oil and petroleum products."""

# fluid, apothecaries
DRAM_FL = (Fraction(1, 8) * FL_OZ).alias("dram_fluid")
"""U.S. apothecaries' fluid dram."""
MINIM = (Fraction(1, 60) * DRAM_FL).alias("minim")
"""U.S. apothecaries' minim."""

# dry, US
BUSHEL = (Decimal("2150.42") * CU_IN).alias("bushel")
"""U.S. dry bushel."""  # H44 C-26
BUSHEL_HEAPED = (Decimal("2747.715") * CU_IN).alias("bushel_heaped")
"""U.S. heaped bushel."""  # H44 C-26
PECK = (Fraction(1, 4) * BUSHEL).alias("peck")
"""U.S. dry peck."""
QUART_DRY = (Fraction(1, 8) * PECK).alias("quart_dry")
"""U.S. dry quart."""
PINT_DRY = (Fraction(1, 2) * QUART_DRY).alias("pint_dry")
"""U.S. dry pint."""
BBL_DRY = (7056 * CU_IN).alias("bbl_dry")
"""U.S. standard barrel for fruits, vegetables and dry commodities
(excluding cranberries)."""  # H44 C-26
BBL_CRANBERRY = (5826 * CU_IN).alias("bbl_cranberry")
"""U.S. standard barrel for cranberries."""  # H44 C-26

# british imperial [WMA1985 1VI]
GAL_IMP = (Decimal("4.54609") * L).alias("gallon_imperial")
"""British Imperial gallon."""
BUSHEL_IMP = (8 * GAL_IMP).alias("bushel_imperial")
"""British Imperial bushel."""
PECK_IMP = (2 * GAL_IMP).alias("peck_imperial")
"""British Imperial peck."""  # H44 C-8
QUARTER_IMP = (8 * BUSHEL_IMP).alias("quarter_imperial")
"""British Imperial quarter."""  # H44 C-8
QUART_IMP = (Fraction(1, 4) * GAL_IMP).alias("quart_imperial")
"""British Imperial quart."""
PINT_IMP = (Fraction(1, 2) * QUART_IMP).alias("pint_imperial")
"""British Imperial pint."""
FL_OZ_IMP = (Fraction(1, 160) * GAL_IMP).alias("ounce_fluid_imperial")
"""British Imperial fluid ounce."""
DRACHM_FL_IMP = (Fraction(1, 8) * FL_OZ_IMP).alias("drachm_fluid_imperial")
"""British Imperial fluid drachm."""
SCRUPLE_FL_IMP = (Fraction(1, 3) * DRACHM_FL_IMP).alias(
    "scruple_fluid_imperial"
)
"""British Imperial fluid scruple."""  # H44 C-8
MINIM_IMP = (Fraction(1, 20) * SCRUPLE_FL_IMP).alias("minim_imperial")
"""British Imperial minim."""  # H44 C-8

# other
ACRE_FOOT = (ACRE * FT).alias("acre_foot")
"""Volume of water that covers one acre to a depth of one foot."""  # H44 C-19
CORD = (128 * FT**3).alias("cord")
"""Cord, a unit of volume for firewood."""  # H44 C-26
WATER_TON = (224 * GAL_IMP).alias("water_ton")
"""Water ton, an English unit of volume,
approximately the volume of a long ton of water."""  # H44 C-28


#
# mass
#

# avoirdupois (common) [H44 C-7]
LB = (Decimal("0.45359237") * KG).alias("pound")
"""Avoirdupois pound mass."""  # H44 C-29
GRAIN = (Fraction(1, 7000) * LB).alias("grain")
"""Grain. Equivalent across Avoirdupois, Troy, and Apothecaries' systems."""  # H44 C-29
OZ = (Fraction(1, 16) * LB).alias("ounce")
"""Avoirdupois ounce."""
DRAM = (Fraction(1, 16) * OZ).alias("dram")
"""Avoirdupois dram."""
CWT = (100 * LB).alias("hundredweight")
"""Short hundredweight (also known as cental)"""
TON = (2000 * LB).alias("ton")
"""Short or net ton."""

# british imperial [WMA1985 1VI]
STONE = (14 * LB).alias("stone")
"""Stone, a British unit of mass."""
QUARTER = (28 * LB).alias("quarter")
"""Quarter, a British unit of mass."""
CWT_LONG = (112 * LB).alias("hundredweight_long")
"""Long hundredweight."""
TON_LONG = (2240 * LB).alias("ton_long")
"""Long, gross or shipper's ton."""

# troy (for precious metals) [H44 C-7]
LB_T = (5760 * GRAIN).alias("pound_troy")
"""Troy pound mass."""
OZ_T = (Fraction(1, 12) * LB_T).alias("ounce_troy")
"""Troy ounce."""
DWT = (Fraction(1, 20) * OZ_T).alias("pennyweight")
"""Pennyweight."""

# apothecaries' (for medicine) [H44 C-8]
LB_AP = (5760 * GRAIN).alias("pound_apothecaries")
"""Apothecaries' pound mass."""
OZ_AP = (Fraction(1, 12) * LB_AP).alias("ounce_apothecaries")
"""Apothecaries' ounce."""
DRAM_AP = (Fraction(1, 8) * OZ_AP).alias("dram_apothecaries")
"""Apothecaries' dram."""
SCRUPLE = (Fraction(1, 3) * DRAM_AP).alias("scruple")
"""Apothecaries' scruple."""
# assaying and gemstones
CARAT = (200 * (MILLI * G)).alias("carat")
"""Metric carat, for gemstones."""  # H44 C-29
POINT_MASS = (Fraction(1, 100) * CARAT).alias("point_mass")
"""Point, for gemstones."""  # H44 C-30
ASSAY_TON = (Decimal("29.167") * G).alias("assay_ton")
"""Assay ton. The mass in milligrams of precious metal from one assay ton of ore
gives the troy ounces per short ton."""  # H44 C-29
QUINTAL = (100 * KG).alias("quintal")
"""Quintal, a historical unit of mass, now usually 100 kg."""  # WMA1985 1VI

#
# misc / engineering
#

# linear velocity
KNOT = (NMI * HOUR**-1).alias("knot")
"""Knot, one nautical mile per hour."""
MPH = (MI * HOUR**-1).alias("mph")
"""Miles per hour."""

# acceleration
FT_PERS2 = FT * S**-2
"""Feet per second squared."""
# force
LBF = (CONST_STANDARD_GRAVITY * LB * M_PERS2).alias("lbf")
"""Pound-force."""
POUNDAL = (LB * FT * S**-2).alias("poundal")
"""Poundal, the force required to accelerate 1 lb by 1 ft/s²."""
KIP = (1000 * LBF).alias("kip")

# pressure [H44 C-59]
PSI = (LBF * IN**-2).alias("psi")
"""Pound-force per square inch, a unit of pressure."""
KSI = (1000 * PSI).alias("ksi")
"""Kilo-pound-force per square inch, a unit of pressure."""
PSF = (LBF * SQ_FT**-1).alias("psf")
"""Pound-force per square foot, a unit of pressure."""
INHG = (LazyProduct((CONST_DENSITY_HG, CONST_STANDARD_GRAVITY)) * IN).alias(
    "inch_of_hg"
)
"""Inch of mercury, a unit of pressure."""  # SP811 B.8
INH2O = (LazyProduct((CONST_DENSITY_H2O, CONST_STANDARD_GRAVITY)) * IN).alias(
    "inch_of_h2o"
)
"""Inch of water (conventional), a unit of pressure."""
INH2O_4C = (Decimal("249.082") * PA).alias("inch_of_water_4c")  # approx
"""Inch of water at 4 °C (temperature of maximum water density)."""
INH2O_60F = (Decimal("248.84") * PA).alias("inch_of_water_60f")  # approx
"""Inch of water at 60 °F."""

SLUG = (LBF * FT_PERS2**-1).alias("slug")
"""A unit of mass that accelerates by 1 ft/s² when a force of 1 lbf is exerted on it."""

# energy
FT_LBF = (FT * LBF).alias("foot_pound")
"""Foot-pound, a unit of energy or work."""
# [SP811 B.8 (footnote 10), H44 C-59]
CAL_IT = (Decimal("4.1868") * J).alias("calorie_it")
"""Calorie (International Table)."""
CAL_TH = (Decimal("4.184") * J).alias("calorie_th")
"""Calorie (thermochemical)."""
CAL_MEAN = (Decimal("4.19002") * J).alias("calorie_mean")  # approx
"""Calorie (mean). The heat required to raise 1 g of water from 0 °C to 100 °C,
divided by 100."""
CAL_15C = (Decimal("4.18580") * J).alias("calorie_15c")  # approx
"""Calorie (at 15 °C). The heat required to raise 1 g of water from 14.5 °C to
15.5 °C."""
CAL_20C = (Decimal("4.18190") * J).alias("calorie_20c")  # approx
"""Calorie (at 20 °C). The heat required to raise 1 g of water from 19.5 °C to
20.5 °C."""
# [SP811 B.8 (footnote 9), H44 C-57]
BTU_IT = (Decimal("1055.05585262") * J).alias("btu_it")
"""British thermal unit (International Table). The most widely used definition."""
BTU_TH = (Decimal("1054.350") * J).alias("btu_th")  # approx
"""British thermal unit (thermochemical)."""
BTU_MEAN = (Decimal("1055.87") * J).alias("btu_mean")  # approx
"""British thermal unit (mean, from 32 °F to 212 °F, divided by 180)."""
BTU_39F = (Decimal("1059.67") * J).alias("btu_39f")  # approx
"""British thermal unit (at 39 °F)."""
BTU_59F = (Decimal("1054.80") * J).alias("btu_59f")  # approx
"""British thermal unit (at 59 °F). Used for American natural gas pricing."""
BTU_60F = (Decimal("1054.68") * J).alias("btu_60f")  # approx
"""British thermal unit (at 60 °F)."""
QUAD = (10**15 * BTU_IT).alias("quad")
"""Quad (International Table). Used by U.S. Department of Energy."""  # H44 B.8 footnote 9

# power [SP811 B.8]
HP = (33000 * (FT * LBF * MIN**-1)).alias("horsepower")
"""Mechanical horsepower (imperial).
See: https://en.wikipedia.org/wiki/Horsepower#Imperial_horsepower"""
HP_METRIC = (75 * (KGF * M * S**-1)).alias("horsepower_metric")
"""Metric horsepower.
See: <https://en.wikipedia.org/wiki/Horsepower#Metric_horsepower_(PS,_KM,_cv,_hk,_pk,_k,_ks,_ch)>"""
HP_BOILER = (Decimal("9809.50") * W).alias("horsepower_boiler")  # approx
"""Boiler horsepower."""
HP_ELECTRIC = (746 * W).alias("horsepower_electric")
"""Electrical horsepower."""
HP_UK = (Decimal("745.70") * W).alias("horsepower_uk")  # approx
"""UK horsepower."""
HP_WATER = (Decimal("746.043") * W).alias("horsepower_water")  # approx
"""Water horsepower."""

#
# misc
#
MPG = (MI * GAL**-1).alias("miles_per_gallon")
"""Miles per gallon, a unit of fuel economy."""  # SP811 B.5
# NOTE: not defining R-value (°F·ft²·h/Btu), U-factor (Btu/(h·ft²·°F)), K-value (Btu·in/(h·ft²·°F))



from ._core import Dimensionless, Log
from ._iso80000 import ACTIVITY_OF_SOLUTE

ACTIVITY_HPLUS = ACTIVITY_OF_SOLUTE["H+"]
ACTIVITY_AMINUS = ACTIVITY_OF_SOLUTE["A-"]
ACTIVITY_HA = ACTIVITY_OF_SOLUTE["HA"]
PH = (-1 * Log(ACTIVITY_HPLUS, base=10)).alias("pH")
KA = Dimensionless("Ka")
"""Acid dissociation constant, a measure of the strength of an acid in solution."""
PKA = (-1 * Log(KA, base=10)).alias("pKa")



from __future__ import annotations

import decimal
import math
import sys
from collections.abc import (
    Generator,
    Hashable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
)
from dataclasses import dataclass, fields, replace
from decimal import Decimal
from fractions import Fraction
from functools import cache
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Final,
    Literal,
    NamedTuple,
    Protocol,
    SupportsAbs,
    SupportsFloat,
    Union,
    final,
    get_args,
    get_origin,
    get_type_hints,
    overload,
    runtime_checkable,
)

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias

    slots: dict[str, bool] = {}
else:
    from typing import TypeAlias

    slots: dict[str, bool] = {"slots": True}

if sys.version_info < (3, 11):
    from typing import NoReturn

    def assert_never(arg: NoReturn) -> NoReturn:
        raise AssertionError("expected code to be unreachable")
else:
    from typing import assert_never

if TYPE_CHECKING:
    from ._fmt import Formatter, _FormatSpec


@runtime_checkable
class SupportsDecimal(SupportsFloat, SupportsAbs[object], Protocol):
    def to_decimal(self, ctx: decimal.Context) -> Decimal: ...


ExprKind: TypeAlias = Literal["dimensionless", "unit", "dimension"]
Number: TypeAlias = Union[SupportsDecimal, Decimal, Fraction, float, int]
_ARGS_NUMBER = get_args(Number)
Name: TypeAlias = str
"""A unique slug to identify the expression. This is used by the default basic
formatter to display the expression and hence should not contain any spaces
to avoid ambiguity. For example, `meter`, `newton`, `reynolds`"""


@runtime_checkable
class Named(Protocol):
    name: Name


# NOTE: these mixins are separate from ExprBase because:
# - the `Alias` expression cannot wrap itself
# - all expressions are formattable, but in the future we also want `QtyKind`
#   (not an ExprBase) to be formattable.


class AliasMixin:
    def alias(self, name: Name, *, allow_prefix: bool = False) -> Aliased:
        """Wrap this expression with a name.

        :param name: Name of this alias, e.g. `newton`
        :param allow_prefix: Whether to allow [prefixes][isqx.Prefix] to be
        attached. This should only be true for some units like `liter`
        """
        return Aliased(self, name=name, allow_prefix=allow_prefix)  # type: ignore


class FormatMixin:
    def __format__(self, fmt: _FormatSpec | str | Formatter) -> str:
        from ._fmt import fmt as format_

        return format_(self, formatter=fmt)  # type: ignore

    def __str__(self) -> str:
        return self.__format__("basic")


class ExprBase(FormatMixin):
    """A base class for a "unit-like" expression.

    It may be of the following forms:

    | Type                                       | Example                   |
    | ------------------------------------------ | ------------------------- |
    | [dimensionless number][isqx.Dimensionless]¹ | `Reynolds`, `Prandtl`     |
    | [base dimension][isqx.BaseDimension]¹       | [length][isqx.DIM_LENGTH]  |
    | [base unit][isqx.BaseUnit]¹                 | [meter][isqx.M]            |
    | [expression raised to a power][isqx.Exp]²   | `Exp(M, 2)`               |
    | [product of expressions][isqx.Mul]²         | `Mul(A, S)`               |
    | [scaled expression][isqx.Scaled]²⁴          | `Scaled(M, 0.3048)`       |
    | [aliased][isqx.Aliased]¹                    | [newton][isqx.N] = kg m s⁻²|
    | [translated expression][isqx.Translated]³   | `Translated(K, -273.15)`  |
    | [logarithmic expression][isqx.Log]¹³        | [dB][isqx.DB]              |
    | [tagged][isqx.Tagged]⁵                      | true vs ground speed      |

    ¹ these expressions are associated with a [name][isqx.NamedExpr].
    ² these expressions can be [aliased with a name][isqx.Aliased].
    ³ these expressions are *terminal*, meaning it cannot be further
      [exponentiated][isqx.Exp], [multiplied][isqx.Mul], [scaled][isqx.Scaled],
      [translated][isqx.Translated] or [aliased][isqx.Aliased] to form a more
      complex expression. However, it can be further [tagged][isqx.Tagged]
      (e.g. surface temperature vs ISA temperature).
    ⁴ can be created by multiplying a [prefix][isqx.Prefix] (e.g. `milli`)
    ⁵ can be created by calling a [quantity kind][isqx.QtyKind] with a unit

    Operator overloading is provided for ergonomic expression construction.

    !!! note
        While dividing expressions is supported, it is strongly discouraged as
        it can lead to operator precedence ambiguity. For example, while Python
        interprets `J / KG / K` as `(J / KG) / K`, it is often clearer to
        represent it as `J * KG**-1 * K**-1`.
    """

    def __pow__(self, exponent: Exponent) -> Exp:
        return Exp(self, exponent)  # type: ignore

    @overload
    def __mul__(self, rhs: Expr) -> Mul: ...

    @overload
    def __mul__(self, rhs: LazyProduct | Number) -> Scaled: ...

    # NOTE: not allowing Prefix as rhs to avoid confusion
    def __mul__(self, rhs: Expr | LazyProduct | Number) -> Mul | Scaled:
        if isinstance(
            rhs, (LazyProduct, SupportsDecimal, Decimal, Fraction, float, int)
        ):
            return Scaled(self, rhs)  # type: ignore
        # make sure KG * M * S becomes flat, not Mul((Mul((KG, M)), S))
        terms_self = self.terms if isinstance(self, Mul) else (self,)
        terms_other = rhs.terms if isinstance(rhs, Mul) else (rhs,)
        return Mul(tuple([*terms_self, *terms_other]))

    def __rmul__(self, lhs: LazyProduct | Number | Prefix) -> Scaled:
        if isinstance(lhs, Prefix):
            return lhs.mul(self)  # type: ignore
        return Scaled(self, lhs)  # type: ignore

    @overload
    def __truediv__(self, rhs: Expr) -> Mul: ...

    @overload
    def __truediv__(self, rhs: LazyProduct | Number) -> Scaled: ...

    def __truediv__(self, rhs: Expr | LazyProduct | Number) -> Mul | Scaled:
        if not isinstance(
            rhs, (LazyProduct, SupportsDecimal, Decimal, Fraction, float, int)
        ):
            return self * rhs**-1

        return Scaled(
            self,  # type: ignore
            LazyProduct(tuple(f for f in _products_inverse(rhs))),
        )  # M / 2 => Scaled(M, Fraction(1, 2))

    def __getitem__(self, tags: tuple[Tag, ...] | Tag) -> Tagged:
        """Attach tags to this expresson."""
        t = tags if isinstance(tags, tuple) else (tags,)
        if isinstance(self, Tagged):
            return Tagged(self.reference, self.tags + t)
        return Tagged(self, t)  # type: ignore


# all expressions should be immutable. helper functions like `kind`, `dimension`
# simplify etc. rely on this fact to cache results.
@dataclass(frozen=True, **slots)
class Dimensionless(Named, ExprBase):
    name: Name
    """Name for the dimensionless number, e.g. `reynolds`, `prandtl`"""


@dataclass(frozen=True, **slots)
class BaseDimension(Named, ExprBase):
    name: Name
    """Name for the base dimension, e.g. `L`, `M`, `T`"""


@dataclass(frozen=True, **slots)
class BaseUnit(Named, ExprBase):
    _dimension: BaseDimension
    """Reference to the base dimension"""
    name: Name
    """Name for the unit, e.g. `m`, `kg`, `s`"""


Exponent: TypeAlias = Union[int, Fraction]
"""An exponent, generally small integers, which can be positive, negative,
or a fraction, but not zero"""


@dataclass(frozen=True, **slots)
class Exp(AliasMixin, ExprBase):
    """An expression raised to an exponent.
    For example, `BaseUnit("meter", Dimension("L")), 2)` is m².
    Can be recursively nested, e.g. `Exp(Exp(METER, 2), Fraction(1, 2))`
    """

    base: _ComposableExpr
    exponent: Exponent
    """Exponent. Avoid using zero to represent dimensionless numbers: 
    use [`isqx.Dimensionless`][] with a name instead."""

    def __post_init__(self) -> None:
        if not isinstance(self.exponent, (int, Fraction)):
            raise CompositionError(
                outer=Exp,
                inner=self.exponent,
                msg=(
                    "exponent must be an integer or a fraction, "
                    f"not {type(self.exponent).__name__}."
                ),
            )
        if self.exponent == 0:
            raise CompositionError(
                outer=Exp,
                inner=self.exponent,
                msg="exponent must not be zero.",
                help="use `Dimensionless` to represent a dimensionless quantity.",
            )
        ref = _unwrap_tagged_or_aliased(self.base)
        if isinstance(ref, Translated):
            raise CompositionError(
                outer=Exp,
                inner=ref,
                msg="translated units (like ℃) are terminal and cannot be exponentiated.",
                help=(
                    "did you mean to exponentiate its "
                    f"absolute reference `{ref.reference}` instead?"
                ),
            )  # prevent ℃². J ℃⁻¹ should be written as J K⁻¹


@dataclass(frozen=True, **slots)
class Mul(AliasMixin, ExprBase):
    """Products of powers of an expression."""

    terms: tuple[_ComposableExpr, ...]
    """A tuple of expressions to be multiplied, preserving the order."""

    def __post_init__(self) -> None:
        if not self.terms:
            raise CompositionError(
                outer=Mul,
                inner=self.terms,
                msg="`Mul` terms must not be empty.",
                help="use `Dimensionless` to represent a dimensionless quantity.",
            )
        current_kind: ExprKind | None = None
        for term in self.terms:
            ref = _unwrap_tagged_or_aliased(term)
            if isinstance(ref, Translated):
                raise CompositionError(
                    outer=Mul,
                    inner=ref,
                    msg="`Translated` units (like ℃) are terminal and cannot be part of a product.",
                    help=f"use its absolute reference `{ref.reference}` instead.",
                )  # prevent ℃ * ℃
            k = kind(term)
            if current_kind is None and k != "dimensionless":
                current_kind = k
            elif current_kind != k and k != "dimensionless":
                raise MixedKindError(terms=self.terms)  # prevent time * seconds


@dataclass(frozen=True, **slots)
class Scaled(AliasMixin, ExprBase):
    reference: BaseUnit | Exp | Mul | Scaled | Aliased | Tagged | Log
    """The unit or dimension that this unit or dimension is based on."""
    factor: Number | LazyProduct | Prefix
    """The exact factor to multiply to this unit to convert it to the reference.
    For example, `1 ft = 0.3048 m`, so the factor is 0.3048.
    """

    def __post_init__(self) -> None:
        if not isinstance(
            self.factor,
            (
                LazyProduct,
                SupportsDecimal,
                Decimal,
                Fraction,
                float,
                int,
                Prefix,
            ),
        ):
            raise CompositionError(
                outer=Scaled,
                inner=self.factor,
                msg=f"factor must be a number, not {type(self.factor).__name__}.",
            )
        ref = _unwrap_tagged_or_aliased(self.reference)
        if isinstance(ref, Translated):
            raise CompositionError(
                outer=Scaled,
                inner=self.reference,
                msg=f"`{type(self.factor).__name__} cannot be scaled.",
            )  # prevent 13 * ℃
        # TODO: prevent BaseDimension from being scaled


@dataclass(frozen=True, **slots)
class Log(AliasMixin, ExprBase):
    """The logarithm of a dimensionless expression [ISO 80000-2:2019 2-13.4]."""

    reference: Dimensionless | Tagged
    """A dimensionless expression"""
    # NOTE: while we should support Exp: log(a^b) = b * log(a), the latter is actually the more "simple" version
    # NOTE: we should also support Mul: log(a * a**-1) is fine, but again we must simplify before we know if its really dimensionless.
    base: Number
    """The base of the logarithm, e.g. 10 for bel, e for neper."""

    def __post_init__(self) -> None:
        ref = self.reference
        is_valid_ref = isinstance(ref, Dimensionless) or (
            isinstance(ref, Tagged) and isinstance(ref.reference, Dimensionless)
        )
        if not is_valid_ref:
            raise CompositionError(
                outer=Log,
                inner=ref,
                msg="`Log` can only wrap a `Dimensionless` expression",
                help="use the `isqx.ratio()` helper for logarithmic units like dB.",
            )


@dataclass(frozen=True, **slots)
class Aliased(Named, ExprBase):
    """An alias for an expression, used to give a more readable name.

    Note that unlike a [tagged][isqx.Tagged] expression,
    [simplification][isqx.simplify] will effectively elide this class.
    """

    reference: Exp | Mul | Scaled | Tagged | Log
    """Expression to be aliased, e.g. `Mul((KG, M, Exp(S, -2)))`"""
    name: Name
    """Name of this alias, e.g. `newton`"""
    allow_prefix: bool = False
    """Whether to allow [prefixes][isqx.Prefix] to be attached.
    This should only be true for some units like `liter`."""

    def __post_init__(self) -> None:
        if not isinstance(
            ref := self.reference, (Exp, Mul, Scaled, Tagged, Log)
        ):
            raise CompositionError(
                outer=Aliased,
                inner=ref,
                msg="`Aliased` can only wrap an `Exp`, `Mul`, `Scaled`, `Tagged`, or `Log` expression.",
            )


@dataclass(frozen=True, **slots)
class Translated(Named, ExprBase):
    """An expression offsetted from some reference unit."""

    reference: BaseUnit | Scaled | Aliased | Tagged
    """The expression that this expression is based on (e.g., `K` for `DEGC`)"""
    offset: Number
    """The exact offset to add to the reference to get this unit.
    For example, `℃ = K - 273.15`, so the offset is -273.15."""
    name: Name

    def __post_init__(self) -> None:
        ref = _unwrap_tagged_or_aliased(self.reference)
        if isinstance(ref, Translated):
            raise CompositionError(
                outer=Translated,
                inner=ref,
                msg="nesting `Translated` expressions is not allowed.",
            )
        if not isinstance(ref, (BaseUnit, Scaled)):
            raise CompositionError(
                outer=Translated,
                inner=ref,
                msg="`Translated` must have a `BaseUnit` or `Scaled` expression as its reference.",
            )


@dataclass(frozen=True, **slots)
class Tagged(AliasMixin, ExprBase):
    """An expression decorated with one of more semantic context tag.

    Similar to how:

    - `Annotated[T, M1, M2, ...]` attaches metadata to some type `T`,
    - `expr[C1, C2, ...]` attaches context to an [expression][isqx.Expr].

    This allows one to "disambiguate" between quantities that share the same
    physical dimension, but have different meanings, e.g.
    geopotential altitude vs. geometric altitude.
    """

    reference: _TaggedAllowedExpr
    tags: tuple[Tag, ...]

    def __post_init__(self) -> None:
        if isinstance(self.reference, Tagged):
            raise CompositionError(
                outer=Tagged,
                inner=self.reference,
                msg="nesting `Tagged` expressions is not allowed. use the `unit[tags]` syntax.",
            )
        for tag in self.tags:
            if isinstance(tag, HasTagValidation):
                tag.__validate_tag__(self.reference, self.tags)


@runtime_checkable
class HasTagValidation(Hashable, Protocol):
    def __validate_tag__(
        self,
        reference: Expr,
        tags: tuple[Tag, ...],
    ) -> None:
        """Check that this tag can be applied to the given expression.

        For example, this can be used to ensure:

        - `decibel` (log level) with log reference unit `voltage` but not
        - `reynolds number` with log reference unit `voltage`.

        :param reference: The expression to apply the tags to.
        :param tags: The tags being applied to the expression.
            This can be used to enforce complex rules (e.g. no duplicates)
        :raises CompositionError: if the tag cannot be applied to the expression
        """
        ...


# using sealed unions instead of ExprBase to facilitate static type checking
Expr: TypeAlias = Union[
    Dimensionless,
    BaseDimension,
    BaseUnit,
    Exp,
    Mul,
    Scaled,
    Aliased,
    Translated,
    Log,
    Tagged,
]
_ARGS_EXPR = get_args(Expr)
NamedExpr: TypeAlias = Union[
    Dimensionless, BaseDimension, BaseUnit, Aliased, Translated
]
_TaggedAllowedExpr: TypeAlias = Union[
    Dimensionless,
    BaseDimension,
    BaseUnit,
    Exp,
    Mul,
    Scaled,
    Aliased,
    Translated,
    Log,
]  # avoid nesting tags
_ComposableExpr: TypeAlias = Union[
    Dimensionless,
    BaseDimension,
    BaseUnit,
    Exp,
    Mul,
    Scaled,
    Aliased,
    Tagged,
    Log,  # dB/m and dB/Hz should be allowed, though with extra care in conversion
]  # avoid terminal (translated) from being further composed
PhysicalUnit: TypeAlias = Union[
    BaseUnit, Exp, Mul, Scaled, Aliased, Tagged
]  # for use in relative

#
# other objects that are not expressions but key to the system
#


@dataclass(frozen=True, **slots)
class Prefix:
    """A prefix, which when multiplied by a [base unit][isqx.BaseUnit] or
    [aliased unit][isqx.Aliased], returns a [scaled unit][isqx.Scaled].

    Note that this is not an [`isqx.Expr`][].
    """

    value: Number
    name: str
    """Name of this prefix, e.g. `milli`, `kibi`"""

    def mul(self, rhs: BaseUnit | Aliased | Tagged) -> Scaled:
        if not isinstance(rhs, (BaseUnit, Aliased, Tagged)):
            raise CompositionError(
                outer=Scaled,
                inner=rhs,
                msg=f"prefixes cannot be applied to `{type(rhs).__name__}`.",
            )  # this will prevent double prefixing

        if isinstance(rhs, BaseUnit):
            if rhs.name == "kilogram":
                raise CompositionError(
                    outer=Scaled,
                    inner=rhs,
                    msg="cannot prefix `kilogram`.",
                    help="apply the prefix to `gram` instead.",
                )
        elif isinstance(rhs, Aliased):
            if not rhs.allow_prefix:
                raise CompositionError(
                    outer=Scaled,
                    inner=rhs,
                    msg=f"The aliased unit `{rhs.name}` does not allow prefixes.",
                )
            if self.name == "kilo" and rhs.name == "gram":
                raise CompositionError(
                    outer=Scaled,
                    inner=rhs,
                    msg="cannot apply prefix `kilo` to `gram`.",
                    help="use the `KG` unit directly.",
                )
        elif isinstance(rhs, Tagged) and not isinstance(
            (ref := _unwrap_tagged_or_aliased(rhs)),
            (BaseUnit, Aliased),
        ):
            raise CompositionError(
                outer=Scaled,
                inner=rhs,
                msg="prefixes cannot be applied to this type of tagged expression.",
                help=f"the inner reference is `{type(ref).__name__}`, which cannot be prefixed.",
            )
        # TODO: robustly handle Scaled(Log())
        return Scaled(rhs, self)


@dataclass(frozen=True, **slots)
class QtyKind:
    r"""An abstract *kind of quantity* (ISO 80000-1) represents a "concept" (e.g.
    speed) *without* a specific unit tied to it.

    When called with a unit, it becomes a [concrete unit with tagged
    context][isqx.Tagged].
    """

    unit_si_coherent: _TaggedAllowedExpr
    """The coherent SI unit (i.e. no conversion factors involved, e.g.
    `M_PERS` for speed, not `KM_PERHOUR`)"""
    tags: tuple[Tag, ...] | None = None

    def si_coherent(self) -> Expr:
        """Return the SI coherent unit with tags."""
        if self.tags is None:
            return self.unit_si_coherent
        return self.unit_si_coherent[self.tags]

    def __getitem__(self, tags: tuple[Tag, ...] | Tag) -> QtyKind:
        """Attach additional tags to this quantity kind."""
        t = tags if isinstance(tags, tuple) else (tags,)
        if self.tags is not None:
            t = self.tags + t
        return replace(self, tags=t)

    def __call__(self, unit: _TaggedAllowedExpr) -> Expr:
        """Create a tagged unit from this quantity kind."""
        if unit is self.unit_si_coherent:
            return self.si_coherent()

        dim_unit = dimension(simplify(unit))
        dim_unit_self = dimension(simplify(self.unit_si_coherent))
        if dim_unit != dim_unit_self:
            raise UnitKindMismatchError(self, unit, dim_unit_self, dim_unit)
        if self.tags is None:
            return unit
        return unit[self.tags]


#
# special tags
# mathematical concepts [ISO 80000-2]
#


@dataclass(frozen=True, **slots)
class _Delta(HasTagValidation):
    """A tag to indicate the quantity as an interval or a difference, (a vector
    in affine space), rather than a particular point on the scale.

    It is independent of a specific origin alone, but when combined with
    [`isqx.OriginAt`][] it can represent a "deviation relative to a specific
    point on the scale".
    """

    kind: Literal["finite", "differential", "inexact_differential"] = "finite"

    def __validate_tag__(self, reference: Expr, tags: tuple[Tag, ...]) -> None:
        _check_duplicate_tags(self, tags)

    # while frozen=True adds the hash, we define our own __hash__ because:
    # - py310's abc.Hashable checks before the dataclass decorator is applied
    # - mypy doesn't understand that frozen=True creates __hash__ dynamically
    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.kind))


DELTA = _Delta("finite")
r"""A tag to indicate a finite change in a quantity $\Delta Q$.

Example: [duration in time][isqx.DURATION]"""
DIFFERENTIAL = _Delta("differential")
r"""A tag to indicate an exact differential $dQ$ (whose integral is
path-independent).

Example: [volume element][isqx.VOLUME_ELEMENT]"""
INEXACT_DIFFERENTIAL = _Delta("inexact_differential")
r"""A tag to indicate an inexact differential $\delta Q$ (whose integral is
path-dependent).

Example: [inexact differential in heat][isqx.INEXACT_DIFFERENTIAL_HEAT]"""


# TODO: Substance (water), Solute, Solution
# TODO: StateOfMatter (solid, liquid, gas, plasma, different allotropes...)
# TODO: vacuum?


@dataclass(frozen=True, **slots)
class Quantity:
    """A simple data container for a value and its unit, for use in
    specifying the [origin of a point on the scale][isqx.OriginAt].

    !!! warning
        This is not an [`isqx.Expr`][] should not be further composed.
    """

    value: Number | LazyProduct
    unit: PhysicalUnit


@dataclass(frozen=True, **slots)
class OriginAt(HasTagValidation):
    """A tag to specify the origin for a point on the scale. Commonly combined
    with [`isqx.DELTA`][] to also represent a change in the quantity.

    Examples:

    - Unix epoch: `S[OriginAt("unix epoch")]`
    - difference in actual temperature and ISA temperature:
      `K[DELTA, OriginAt((15, CELSIUS))]`
    - difference in pressure: `Pa[DELTA, OriginAt((100, PSI))]`
    """

    location: Quantity | Hashable
    """The location of the "zero point" of the measurement.

    Can be a quantity (value + unit, e.g. `(100, PSI)`) or simply a hashable
    object (e.g. the string "unix epoch"). Hashable objects are useful for
    scenarios where the origin itself is contextual, for example:

    - the height above the ground is dependent on the geographic location
    - the stock price relative to yesterday's close
    - time elapsed relative to the engine ignition
    """

    def __validate_tag__(self, reference: Expr, tags: tuple[Tag, ...]) -> None:
        if (
            isinstance(q := self.location, Quantity)
            and not isinstance(reference, Dimensionless)
            and q.unit != reference
        ):
            raise CompositionError(
                outer=OriginAt,
                inner=reference,
                msg=(
                    f"expression {reference} is not compatible with "
                    f"`OriginAt.location` of unit {q.unit}"
                ),
                help=f"`OriginAt.location` must be of unit {reference}",
            )
        _check_duplicate_tags(self, tags)

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.location))


# TODO: add tag `AtCondition` (Delta pressure = 0, temperature = 293.15 K)
# TODO: add tag `Rotation`? clockwise/counterclockwise
# TODO: add tag `Direction`? into/out of system

_RATIO = Dimensionless("ratio")


def ratio(numerator: Expr, denominator: Expr | Quantity) -> Tagged:
    """Returns a dimensionless between two expressions of the same dimension.

    The denominator can be a [quantity (value + unit)][isqx.Quantity] to
    represent logarithmic units like [`isqx.DBW`][].

    See: https://en.wikipedia.org/wiki/Dimensionless_quantity#Ratios,_proportions,_and_angles"""
    return _RATIO[_RatioBetween(numerator, denominator)]


@dataclass(frozen=True, **slots)
class _RatioBetween(HasTagValidation):
    numerator: Expr
    denominator: Expr | Quantity

    def __post_init__(self) -> None:
        # TODO: enforce dimensionless:
        # we abuse numerator * (denominator.value * denominator.unit)**-1
        # and check we indeed get dimensionless or scaled dimensionless.
        pass

    def __validate_tag__(self, reference: Expr, tags: tuple[Tag, ...]) -> None:
        if reference is not _RATIO:
            raise CompositionError(
                outer=_RatioBetween,
                inner=reference,
                msg="`RatioBetween` tag can only be applied to a ratio.",
                help="use the `isqx.ratio()` helper function.",
            )
        _check_duplicate_tags(self, tags)

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.numerator, self.denominator))


@dataclass(frozen=True, **slots)
class _Tensor(HasTagValidation):
    rank: int

    def __validate_tag__(self, reference: Expr, tags: tuple[Tag, ...]) -> None:
        _check_duplicate_tags(self, tags)

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.rank))


# NOTE: `var: Annotated[T, x]` can be interpreted as "the numerical value of
# `var` is a measurement on the scale `x`". If `T` is a float, then it is a
# scalar; if `T` is numpy array, it can be a tensor.
# Use `_Tensor` to indicate that the quantity *can*, but does not have to,
# be a tensor.
VECTOR = _Tensor(rank=1)
"""A tag to indicate that the quantity can be a vector.
[ISO 80000-2:2019 2-18.1]. It is not restricted to three dimensions.

Examples:

- velocity vector: `M_PERS[VECTOR]`
- force vector: `F[VECTOR]`"""
TENSOR_SECOND_ORDER = _Tensor(rank=2)
"""A tag to indicate that the quantity can be a second-order tensor.

Examples:

- stress tensor: `(N * M**-2)["stress", TENSOR_SECOND_ORDER]`"""


@dataclass(frozen=True, **slots)
class _CoordinateSystem(HasTagValidation, Named):
    name: Literal["cartesian", "spherical", "cylindrical"] | str
    # NOTE: allowing arbitrary names for now because different game engines
    # don't agree to one (also ENU, NED...)

    def __validate_tag__(self, reference: Expr, tags: tuple[Tag, ...]) -> None:
        _check_duplicate_tags(self, tags)

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.name))


# coordinate systems, TODO: left handed systems.
CARTESIAN = _CoordinateSystem(name="cartesian")
r"""A tag to indicate that the quantity is expressed in Cartesian coordinates
($x$, $y$, $z$) [ISO 80000-2:2019 2-17.1].

The position vector is $\mathbf{r} = x \mathbf{e}_x + y \mathbf{e}_y +
z \mathbf{e}_z$, and its differential is $d\mathbf{r} = dx \mathbf{e}_x +
dy \mathbf{e}_y + dz \mathbf{e}_z$.
"""
CYLINDRICAL = _CoordinateSystem(name="cylindrical")
r"""A tag to indicate that the quantity is expressed in cylindrical coordinates
(radial distance $\rho$, azimuth $\varphi$, axial coordinate or height $z$)
[ISO 80000-2:2019 2-17.2].

The position vector is $\mathbf{r} = \rho \mathbf{e}_\rho(\varphi) +
z \mathbf{e}_z$, and its differential is $d\mathbf{r} =
d\rho \mathbf{e}_\rho(\varphi) + \rho d\varphi \mathbf{e}_\varphi(\varphi) +
dz \mathbf{e}_z$.

If $z = 0$, it is equivalent to polar coordinates.
"""
SPHERICAL = _CoordinateSystem(name="spherical")
r"""A tag to indicate that the quantity is expressed in spherical coordinates
(radial distance $r$, polar angle $\theta$, azimuthal angle $\varphi$)
[ISO 80000-2:2019 2-17.3].

The position vector is $\mathbf{r} = r \mathbf{e}_r(\theta, \varphi)$,
and its differential is $d\mathbf{r} = dr \mathbf{e}_r(\theta, \varphi) +
r d\theta \mathbf{e}_\theta(\theta, \varphi) + r \sin(\theta) d\varphi
\mathbf{e}_\varphi(\theta, \varphi)$.
"""

# TODO: while it is tempting to add a tag that "constrains" the quantity to be
# greater than zero, much like `annotated_types.Gt`, we need to be very careful
# because some quantities can be negative, e.g.
# negative thermodynamic temperatures or altitudes below sea level.
# https://github.com/mpusz/mp-units/issues/468


@dataclass(frozen=True, **slots)
class _Complex(HasTagValidation):
    def __validate_tag__(self, reference: Expr, tags: tuple[Tag, ...]) -> None:
        _check_duplicate_tags(self, tags)

    def __hash__(self) -> int:
        return hash(self.__class__.__name__)


COMPLEX = _Complex()
"""A tag to indicate that the quantity is complex-valued."""


# iso/iec 80000-specific
@dataclass(frozen=True, **slots)
class PhotometricCondition(HasTagValidation):
    """Photometric condition [ISO 80000-7]. Can be:

    - [Photopic vision (cones in human vision, in daylight)](https://en.wikipedia.org/wiki/Photopic_vision)
    - [Scotopic vision (rods in human vision, at night)](https://en.wikipedia.org/wiki/Scotopic_vision)
    - [Mesopic/twilight vision (rods and cones in human vision)](https://en.wikipedia.org/wiki/Mesopic_vision)
    - CIE 1964 standard colorimetric observer (10° photopic photometric observer)
    - CIE 1988 modified 2° spectral luminous efficiency function for photopic vision
    - Any other custom condition.
    """

    kind: (
        Literal[
            "photopic",
            "scotopic",
            "mesopic",
            "photopic_cie_1964_10degree",
            "photopic_cie_1988_mod_2degree",
        ]
        | str
    )

    def __validate_tag__(self, reference: Expr, tags: tuple[Tag, ...]) -> None:
        _check_duplicate_tags(self, tags)

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.kind))


def _check_duplicate_tags(
    tag: Tag,
    tags: tuple[Tag, ...],
) -> None:
    tag_type = type(tag)
    if sum(1 for t in tags if isinstance(t, tag_type)) > 1:
        raise CompositionError(
            outer=Tagged,
            inner=tag,
            msg=f"tag `{tag_type.__name__}` cannot be applied multiple times.",
        )


Tag: TypeAlias = Union[
    OriginAt,
    _RatioBetween,
    _Delta,
    _Tensor,
    _CoordinateSystem,
    _Complex,
    PhotometricCondition,
    HasTagValidation,
    Hashable,
]
"""A tag can be any hashable object (e.g. frozen dataclasses or strings)."""

#
# this ends the basic expression system used for documentation purposes.
# the rest of the code are *runtime* helpers that transform and manipulate
# the expressions.
#
# the `dimension` and `kind` helper is sometimes used in `__post_init__` to make
# invalid states unrepresentable. this comes with a minor performance
# penalty but is worth it.
#


@overload
def dimension(expr: Dimensionless) -> Dimensionless: ...
@overload
def dimension(expr: BaseDimension) -> BaseDimension: ...
@overload
def dimension(expr: BaseUnit) -> BaseDimension: ...
@overload
def dimension(expr: Exp) -> Exp: ...
@overload
def dimension(expr: Mul) -> Mul: ...
@overload
def dimension(expr: Log) -> Dimensionless: ...
@overload
def dimension(expr: Tagged) -> Tagged | Expr: ...
@overload
def dimension(expr: Aliased | Scaled | Translated) -> Expr: ...
@cache
def dimension(expr: Expr) -> Expr:
    """Return the dimension of this "unit-like" expression.
    Note that it does not perform simplification.

    Examples:

    - `Exp(M, 2)` -> `Exp(DIM_LENGTH, 2)`
    - `Mul(M, Exp(S, -1))` -> `Mul(DIM_LENGTH, Exp(DIM_TIME, -1))`
    """
    if isinstance(expr, (Dimensionless, BaseDimension)):
        return expr
    if isinstance(expr, BaseUnit):
        return expr._dimension
    if isinstance(expr, (Aliased, Scaled, Translated)):
        return dimension(expr.reference)
    if isinstance(expr, Exp):
        return Exp(dimension(expr.base), expr.exponent)  # type: ignore
    if isinstance(expr, Mul):
        return Mul(tuple(dimension(term) for term in expr.terms))  # type: ignore
    if isinstance(expr, Tagged):
        ref_dim = dimension(expr.reference)
        # note: log level's dimension is always dimensionless
        # so we need to strip away tags that dont make any more sense
        tags = []
        for tag in expr.tags:
            if isinstance(tag, HasTagValidation):
                try:
                    tag.__validate_tag__(ref_dim, expr.tags)
                except CompositionError:
                    continue
            tags.append(tag)
        if not tags:
            return ref_dim
        return ref_dim[tuple(tags)]
    if isinstance(expr, Log):
        return Dimensionless(f"log_{repr(expr.reference)}")
    assert_never(expr)


@overload
def kind(expr: Dimensionless | Log) -> Literal["dimensionless"]: ...
@overload
def kind(expr: BaseDimension) -> Literal["dimension"]: ...
@overload
def kind(expr: BaseUnit) -> Literal["unit"]: ...
@overload
def kind(
    expr: Exp | Mul | Scaled | Aliased | Translated | Tagged,
) -> ExprKind: ...
@cache
def kind(expr: Expr) -> ExprKind:
    """Whether this expression is a unit, dimension, or dimensionless."""
    if isinstance(expr, Dimensionless):
        return "dimensionless"
    if isinstance(expr, BaseDimension):
        return "dimension"
    if isinstance(expr, BaseUnit):
        return "unit"
    if isinstance(expr, Exp):
        return kind(expr.base)
    if isinstance(expr, (Scaled, Aliased, Translated, Tagged)):
        return kind(expr.reference)
    if isinstance(expr, Mul):
        # all terms have a consistent underlying kind (unit/dimension)
        for term in expr.terms:
            term_kind = kind(term)
            if term_kind != "dimensionless":
                return term_kind
        return "dimensionless"
        # everything left are dimensionless to the power of something
    if isinstance(expr, Log):
        return "dimensionless"
    assert_never(expr)


#
# simplify
#


SimplifiedExpr: TypeAlias = Union[
    Dimensionless,
    BaseDimension,
    BaseUnit,
    Exp,
    Mul,
    Scaled,
    Translated,
    Log,
    Tagged,
]  # no Aliased


@overload
def simplify(expr: Dimensionless) -> Dimensionless: ...
@overload
def simplify(expr: BaseDimension) -> BaseDimension: ...
@overload
def simplify(expr: BaseUnit) -> BaseUnit: ...
@overload
def simplify(expr: Aliased) -> SimplifiedExpr: ...
@overload
def simplify(expr: Translated) -> Translated: ...
@overload
def simplify(expr: Log) -> Log: ...
@overload
def simplify(expr: Tagged) -> Tagged: ...
@overload
def simplify(
    expr: Exp | Mul | Scaled,
) -> Union[
    Dimensionless,
    BaseDimension,
    BaseUnit,
    Exp,
    Mul,
    Scaled,
    Translated,
]: ...
@cache
def simplify(expr: Expr) -> SimplifiedExpr:
    if isinstance(expr, (Dimensionless, BaseDimension, BaseUnit)):
        return expr
    if isinstance(expr, Aliased):
        return simplify(expr.reference)
    if isinstance(expr, Translated):
        return Translated(
            simplify(expr.reference),  # type: ignore
            expr.offset,
            expr.name,
        )
    if isinstance(expr, Tagged):
        return Tagged(simplify(expr.reference), expr.tags)  # type: ignore
    base_exponent_pairs: dict[SimplifiedExpr, Exponent] = {}
    scaled_conversions: list[tuple[Scaled, Exponent]] = []
    _decompose_expr(expr, 1, base_exponent_pairs, scaled_conversions)
    return _build_canonical_expr(base_exponent_pairs, scaled_conversions)


# migrate to pattern matching when we drop support for py3.9
def _decompose_expr(
    expr: Expr,  # ‾
    exponent: Exponent,  # *
    base_exponent_pairs_mut: MutableMapping[SimplifiedExpr, Exponent],
    scaled_conversions_mut: MutableSequence[tuple[Scaled, Exponent]],  # ^
) -> None:
    """Recursively traverse an expression tree to flatten it.

    This is the first step of the simplification, example:

    - expr=Mul(Scaled(M, 0.3048, "FT"), Exp(S, -1)), exponent=2
        - expr=Scaled(M, 0.3048, "FT"), exponent=2
            - scaled_conversions_mut += [(Scaled(M, 0.3048, "FT"), 2)]
            - expr=M, exponent=2
                - base_exponent_pairs_mut[M] += 2
        - expr=Exp(S, -1), exponent=2
            - expr=Expr(S, -2), exponent=1
                - base_exponent_pairs_mut[S] += -2
    """
    if isinstance(expr, Aliased):
        _decompose_expr(
            expr.reference,
            exponent,
            base_exponent_pairs_mut,
            scaled_conversions_mut,
        )
    elif isinstance(
        expr,
        (Dimensionless, BaseDimension, BaseUnit, Tagged, Translated, Log),
    ):
        # we hit a fundamental-like unit (we treat tagged as unique bases).
        # add its accumulated exponent
        base_exponent_pairs_mut.setdefault(expr, 0)
        base_exponent_pairs_mut[expr] += exponent
    elif isinstance(expr, Exp):
        # (xᵃ)ᵇ -> xᵃᵇ
        # ‾‾‾‾*    ‾**
        _decompose_expr(
            expr.base,
            expr.exponent * exponent,
            base_exponent_pairs_mut,
            scaled_conversions_mut,
        )
    elif isinstance(expr, Mul):
        # (xyᵃ)ᵇ -> xᵇyᵃᵇ
        # ‾‾‾‾‾*    ‾*‾**
        for term in expr.terms:
            _decompose_expr(
                term,
                exponent,
                base_exponent_pairs_mut,
                scaled_conversions_mut,
            )
    elif isinstance(expr, Scaled):
        # (kx)ᵇ -> kᵇxᵇ
        # ‾‾‾‾*    ^^‾*
        scaled_conversions_mut.append((expr, exponent))
        _decompose_expr(
            expr.reference,
            exponent,
            base_exponent_pairs_mut,
            scaled_conversions_mut,
        )
    else:
        assert_never(expr)


__DIMENSIONLESS_SIMPLIFIED: Final[Dimensionless] = Dimensionless(
    name="from_simplified"
)


def _build_canonical_expr(
    base_exponent_pairs: Mapping[SimplifiedExpr, Exponent],
    scaled_conversions: list[tuple[Scaled, Exponent]],
) -> SimplifiedExpr:
    """Construct a canonical expression from flattened parts.

    This is second step of the simpification. Examples:

    - `{}` and `[]` -> dimensionless
    - `{M: 1}` and `[]` -> `M`
    - `{M: 2}` and `[]` -> `Exp(M, 2)`
    - `{M: 1, S: -2}` and `[]` -> `Mul(M, Exp(S, -2))`
    - `{...}` and `[(Scaled(M, 0.3048, "FT"), 2)]` -> result will be wrapped:
      `Scaled(reference=Mul(..., factor=LazyProduct(...)`
    """
    no_conversions_involved = not scaled_conversions
    simplified_expr: Expr
    # remove zero exponents and ensure canonical order by names
    base_exponent_pairs_sorted = sorted(
        filter(lambda item: item[1] != 0, base_exponent_pairs.items()),
        key=lambda item: repr(item[0]),
    )
    if not base_exponent_pairs_sorted:
        simplified_expr = __DIMENSIONLESS_SIMPLIFIED
    elif len(base_exponent_pairs_sorted) == 1:
        base, exponent = base_exponent_pairs_sorted[0]
        simplified_expr = base if exponent == 1 else Exp(base, exponent)  # type: ignore
    else:
        simplified_expr = Mul(
            tuple(
                base if exponent == 1 else Exp(base, exponent)  # type: ignore
                for base, exponent in base_exponent_pairs_sorted
            )
        )
    if no_conversions_involved:
        return simplified_expr

    return Scaled(
        reference=simplified_expr,  # type: ignore
        factor=LazyProduct.from_derived_conversions(scaled_conversions),
    )


# TODO: in the future, we want a as_basis() function that accepts set[Expr]
# why: we sometimes want to re-express some unit not in MKS, but CGS.
# that will require linear algebra to solve the dimensional exponents but
# we're leaving that as optional and far in the future.

#
# converter
#


def _converter_new(
    scale: Number, offset: Number
) -> Converter | NonAffineConverter:
    # unless conversions involve celsius/fahrenheit/log units, non-affine
    # conversions are rare
    if not offset:
        return Converter(scale=scale)
    return NonAffineConverter(scale=scale, offset=offset)


@runtime_checkable
class SupportsConversion(Protocol):
    def __call__(self, value):  # type: ignore
        """Convert a value in the origin unit to the target unit.

        :param value: An `int`, `float`, [fractions.Fraction][],
            or any other numeric type^.

            ^ If the converter was created with `exact=True`, the `scale`
            will be fractions and thus may not be compatible with many numeric
            libraries (e.g. `fractions.Fraction * numpy.array` fails).
            [decimal.Decimal][] inputs should be converted into
            [fractions.Fraction][].
        """


@dataclass(frozen=True, **slots)
class Converter(SupportsConversion):
    scale: Number

    def __call__(self, value):  # type: ignore
        return self.scale * value


@dataclass(frozen=True, **slots)
class NonAffineConverter(SupportsConversion):
    scale: Number
    offset: Number

    def __call__(self, value):  # type: ignore
        return value * self.scale + self.offset


class _LogInfo(NamedTuple):
    k: Number | LazyProduct
    b: Number
    q_measured_unit: PhysicalUnit
    q_ref: Quantity
    other_tags: tuple[Tag, ...]


def _get_log_info(info: _ConversionInfo) -> _LogInfo | None:
    if not isinstance(info.expr, Log):
        return None
    log_expr = info.expr
    if not isinstance(log_expr.reference, Tagged):
        return None
    ratio_between_tag = None
    other_tags: list[Tag] = []
    for tag in log_expr.reference.tags:
        if isinstance(tag, _RatioBetween):
            ratio_between_tag = tag
        else:
            other_tags.append(tag)
    if ratio_between_tag is None or (
        not isinstance(ratio_between_tag.denominator, Quantity)
    ):
        # pH, pKa isn't log(ratio), and
        # generic dB does not refer to a specific reference value
        return None
    # TODO: handle Log(Exp(ratio, 2))
    return _LogInfo(
        k=info.factor,
        b=log_expr.base,
        q_measured_unit=log_expr.reference,
        q_ref=ratio_between_tag.denominator,
        other_tags=tuple(other_tags),
    )


def convert(
    origin: Expr,
    target: Expr,
    *,
    exact: bool = False,
    ctx: decimal.Context | None = None,
) -> Converter | NonAffineConverter:
    """Create a new unit converter from one unit to another.

    Checks that the underlying dimension are compatible
    (e.g. `USD/year` and `HKD/hour`) and computes the total scaling factor.
    """
    ctx = ctx or decimal.getcontext()
    origin_simpl = simplify(origin)
    target_simpl = simplify(target)

    info_origin = _flatten(origin_simpl)
    info_target = _flatten(target_simpl)

    log_info_origin = _get_log_info(info_origin)
    log_info_target = _get_log_info(info_target)

    if log_info_origin and log_info_target:
        if not (
            log_info_origin.b == log_info_target.b
            and log_info_origin.q_measured_unit
            == log_info_target.q_measured_unit
            and log_info_origin.q_ref.unit == log_info_target.q_ref.unit
            and log_info_origin.other_tags == log_info_target.other_tags
        ):
            return _convert_logarithmic(
                log_info_origin, log_info_target, exact=exact, ctx=ctx
            )
        # if the underlying log unit is the same (same base and reference qty type),
        # then it's a simple linear scaling (e.g. Np -> dNp)
    elif log_info_origin or log_info_target:
        raise NonLinearConversionError(
            origin=origin,
            target=target,
        )  # e.g. V = V_ref * b**(L_dBV / k), L_dbV = k * log_b(V / V_ref)

    if (origin_kind := kind(origin_simpl)) != (
        target_kind := kind(target_simpl)
    ):
        raise KindMismatchError(
            origin_kind=origin_kind, target_kind=target_kind
        )

    origin_dim = dimension(info_origin.expr)
    target_dim = dimension(info_target.expr)
    origin_dim_terms = (
        origin_dim.terms if isinstance(origin_dim, Mul) else (origin_dim,)
    )
    target_dim_terms = (
        target_dim.terms if isinstance(target_dim, Mul) else (target_dim,)
    )
    if origin_dim_terms != target_dim_terms:
        raise DimensionMismatchError(origin, target, origin_dim, target_dim)
    # we have:
    #   v_abs = scale_origin * v_origin + offset_origin
    #   v_abs = scale_target * v_target + offset_target
    # then:
    #   v_target = (scale_origin / scale_target) * v_origin +
    #              (offset_origin - offset_target) / scale_target
    scale_origin = list(_products(info_origin.factor))

    inv_scale_target = tuple(f for f in _products_inverse(info_target.factor))
    scale = LazyProduct(tuple([*scale_origin, *inv_scale_target]))

    offset_numerator = _factor_to_fraction(
        info_origin.offset, ctx=ctx
    ) - _factor_to_fraction(info_target.offset, ctx=ctx)
    offset = LazyProduct(tuple([offset_numerator, *inv_scale_target]))

    return _converter_new(
        scale=scale.to_exact(ctx=ctx) if exact else scale.to_approx(),
        offset=offset.to_exact(ctx=ctx) if exact else offset.to_approx(),
    )


def _convert_logarithmic(
    origin_info: _LogInfo,
    target_info: _LogInfo,
    *,
    exact: bool,
    ctx: decimal.Context,
) -> Converter | NonAffineConverter:
    r"""
    With $L_1 = k_1 \log_{b_1}\left(\frac{Q}{Q_{\text{ref}_1}}\right)$,
    $L_2 = k_2 \log_{b_2}\left(\frac{Q}{Q_{\text{ref}_2}}\right)
    = \underbrace{\left(\frac{k_2}{k_1}\frac{\ln b_1}{\ln b_2}\right)}_\text{scale}
    L_1 + \underbrace{k_2 \log_{b_2}\left(\frac{Q_{\text{ref}_1}}
    {Q_{\text{ref}_2}}\right)}_\text{offset}$
    """
    if origin_info.other_tags != target_info.other_tags:
        raise DimensionMismatchError(
            origin_info.q_measured_unit,
            target_info.q_measured_unit,
            dimension(simplify(origin_info.q_measured_unit)),
            dimension(simplify(target_info.q_measured_unit)),
        )

    ref_converter = convert(
        origin=origin_info.q_ref.unit,
        target=target_info.q_ref.unit,
        exact=True,
        ctx=ctx,
    )
    if isinstance(ref_converter, NonAffineConverter):
        raise NonLinearConversionError(
            origin_info.q_ref.unit, target_info.q_ref.unit
        )

    k1 = _factor_to_fraction(origin_info.k, ctx=ctx)
    k2 = _factor_to_fraction(target_info.k, ctx=ctx)
    b1 = origin_info.b
    b2 = target_info.b
    ref_ratio_f = _factor_to_fraction(ref_converter.scale, ctx=ctx)
    k_ratio = k2 / k1

    if exact:
        b1_d = _fraction_to_decimal(_factor_to_fraction(b1, ctx=ctx))
        b2_d = _fraction_to_decimal(_factor_to_fraction(b2, ctx=ctx))
        ref_ratio_d = _fraction_to_decimal(ref_ratio_f)

        ln_b1_d = b1_d.ln(ctx)
        ln_b2_d = b2_d.ln(ctx)
        ln_ref_ratio_d = ref_ratio_d.ln(ctx)

        scale = k_ratio * Fraction(ln_b1_d) / Fraction(ln_b2_d)
        offset = k2 * Fraction(ln_ref_ratio_d) / Fraction(ln_b2_d)
        return _converter_new(scale, offset)
    else:
        ln_b2_fl = math.log(float(b2))
        scale_fl = float(k_ratio) * math.log(float(b1)) / ln_b2_fl
        offset_fl = float(k2) * math.log(float(ref_ratio_f)) / ln_b2_fl
        return _converter_new(scale_fl, offset_fl)


class _ConversionInfo(NamedTuple):
    expr: Expr
    """Absolute (non-translated, non-scaled) reference"""
    factor: Number | LazyProduct
    """Total scaling factor to convert from this unit to the absolute reference"""
    offset: Number | LazyProduct
    """Total offset to convert from this unit to the absolute reference"""
    tag_origin_at: OriginAt | None = None
    tag_delta: _Delta | None = None


def _flatten(
    expr_simpl: SimplifiedExpr,
) -> _ConversionInfo:
    """Recursively flattens an expression into its absolute base unit."""
    if isinstance(
        expr_simpl,
        (BaseUnit, BaseDimension, Dimensionless, Log, Exp, Mul),
    ):
        # since `Exp(Scaled(x, a), b)` → `Scaled(Exp(x, b), a * b)`
        # similarly, `Mul((Scaled(...),))` → `Scaled(Mul((...),),)`
        return _ConversionInfo(expr_simpl, 1, 0)
    elif isinstance(expr_simpl, Scaled):
        # so `Scaled` should always be pushed to outmost level.
        # furthermore, since nested `Scaled(Scaled(x, c), d)` would simplify to
        # `Scaled(x, c * d)`, we don't need to recursively get the factor.
        factor = f.value if isinstance(f := expr_simpl.factor, Prefix) else f
        return _ConversionInfo(expr_simpl.reference, factor, 0)
    elif isinstance(expr_simpl, Tagged):
        info_ref = _flatten(expr_simpl.reference)  # type: ignore
        tag_origin_at = None
        tag_delta = None
        for tag in expr_simpl.tags:
            if isinstance(tag, OriginAt):
                tag_origin_at = tag
            elif isinstance(tag, _Delta):
                tag_delta = tag
        return _ConversionInfo(
            info_ref.expr[expr_simpl.tags],
            info_ref.factor,
            info_ref.offset,
            tag_origin_at=tag_origin_at,
            tag_delta=tag_delta,
        )
    elif isinstance(expr_simpl, Translated):
        info_ref = _flatten(expr_simpl.reference)  # type: ignore
        assert info_ref.offset == 0, (
            f"inner reference of {expr_simpl=} should not have any offset"
        )
        # v_local = v_ref + offset_local
        #   v_abs = v_ref * factor_ref
        #         = v_local * factor_ref - offset_local * factor_ref
        new_offset = LazyProduct(
            (-1, expr_simpl.offset, *_products(info_ref.factor))
        )
        return _ConversionInfo(info_ref.expr, info_ref.factor, new_offset)
    else:
        assert_never(expr_simpl)


#
# utilities
#


def _unwrap_tagged_or_aliased(expr: Expr) -> Expr:
    # `Translated` and `Log` are terminal, this plugs a loophole
    # NOTE: while Aliased(Aliased(...)) and Tagged(Tagged(...)) are disallowed,
    # Aliased(Tagged(...)) is allowed, so we need to unwrap recursively.
    if isinstance(expr, (Tagged, Aliased)):
        return _unwrap_tagged_or_aliased(expr.reference)
    return expr


def _products(
    factor: Number | LazyProduct | Prefix,
) -> Generator[tuple[Number, Exponent] | Number]:
    if isinstance(factor, Prefix):
        yield factor.value
    elif isinstance(factor, LazyProduct):
        for product in factor.products:
            yield product
    else:
        yield factor


def _products_inverse(
    factor: Number | LazyProduct | Prefix,
) -> Generator[tuple[Number, Exponent] | Number]:
    for item in _products(factor):
        if isinstance(item, tuple):
            base, exponent = item
            yield (base, -exponent)
        else:
            yield (item, -1)


def _factor_to_fraction(
    factor: Number | LazyProduct | Prefix, *, ctx: decimal.Context
) -> Fraction:
    if isinstance(factor, Prefix):
        return _factor_to_fraction(factor.value, ctx=ctx)
    if isinstance(factor, LazyProduct):
        return Fraction(factor.to_exact(ctx=ctx))
    elif isinstance(factor, (Decimal, float, int)):
        return Fraction(factor)
    elif isinstance(factor, SupportsDecimal):
        return Fraction(factor.to_decimal(ctx=ctx))
    elif not isinstance(factor, Fraction):
        raise assert_never(factor)
    return factor


def _fraction_to_decimal(fraction: Fraction) -> Decimal:
    return fraction.numerator / Decimal(fraction.denominator)


@dataclass(frozen=True, **slots)
class LazyProduct(SupportsFloat):
    r"""Represents a lazy product of a sequence of numbers raised to an optional
    exponent, i.e. $\prod_i x_i$, or $\prod_i x_i^{e_i}$.

    Lazy evaluation allows the choice between evaluating it to an exact value
    (taking longer to compute, useful for financial calculations) or an
    approximate float.
    """  # TODO: support LazyProduct itself as the base

    products: tuple[tuple[Number, Exponent] | Number, ...]

    @classmethod
    def from_derived_conversions(
        cls,
        derived_conversions: Sequence[tuple[Scaled, Exponent]],
    ) -> LazyProduct:
        products: list[tuple[Number, Exponent] | Number] = []
        for scaled, exponent in derived_conversions:
            for inner_item in _products(scaled.factor):
                if isinstance(inner_item, tuple):
                    base, inner_exp = inner_item
                    products.append((base, inner_exp * exponent))
                else:
                    products.append((inner_item, exponent))
        return cls(tuple(products))

    def to_approx(self) -> float:
        """Reduce it to an approximate float value. Good enough for most
        applications."""
        product = 1.0
        for item in self.products:
            base, exponent = item if isinstance(item, tuple) else (item, 1)
            if base == 0:
                if exponent > 0:
                    return 0.0
                if exponent == 0:
                    continue  # 0 ** 0 = 1
            if base == 1:
                continue
            product *= float(base) ** float(exponent)
        return product

    # NOTE: not defining `__mul__` to avoid confusion.
    def __float__(self) -> float:
        return self.to_approx()

    def to_exact(self, ctx: decimal.Context) -> Fraction | Decimal:
        """Reduce it to an *exact* fraction or decimal.

        :param ctx: The decimal context (precision, rounding, etc.) to use.
            `decimal.getcontext()` can be used to get the current context.

        The return type depends on the items of each product:

        ```
                         +--------+-------+----------------+
                         |  None  |  int  | Fraction(p, q) | <- exponent
        +----------------+--------+-------+----------------+
        | Decimal        |   .    |   .   |       x        |
        | Fraction(a, b) |   .    |   .   |       *        |
        | float          |   .    |   x   |       x        |
        | int            |   .    |   .   |       ^        |
        +----------------+--------+-------+----------------+
          ^
          |_ base

        Can `base ** exponent` be represented exactly a `Fraction`?

        . Yes
        * Only if q > 0 and a^p and b^q are the perfect q-th power of an integer
          e.g. (8/27) ** (1/3) = 2/3
        ^ Sometimes, e.g. 4 ** (1/2)
        x No, decimal only
        ```

        For simplicity, only cases that can definitively be represented as a
        `Fraction` are returned as such. A `Decimal` is returned otherwise.
        """
        # accumulate products in two streams: if the "tripwire" for decimal
        # is hit, we must return Decimal.
        product_fraction = Fraction(1)
        product_decimal = Decimal(1)
        for item in self.products:
            if not isinstance(item, tuple):  # no exponent
                product_fraction *= _factor_to_fraction(item, ctx=ctx)
                continue
            base, exponent = item
            if base == 0:
                if exponent > 0:
                    return Fraction(0)
                if exponent < 0:
                    raise ZeroDivisionError
                continue  # 0 ** 0 = 1
            if base == 1:
                continue
            if isinstance(exponent, int):
                # most of the time, we can represent it as Fraction
                if isinstance(base, float):
                    base_decimal = ctx.create_decimal_from_float(base)
                    product_decimal *= base_decimal**exponent
                else:
                    base_fraction = _factor_to_fraction(base, ctx=ctx)
                    product_fraction *= base_fraction**exponent
            elif isinstance(exponent, Fraction):
                # but raising to a Fraction exponent requires decimal
                if isinstance(base, SupportsDecimal):
                    base_decimal = base.to_decimal(ctx=ctx)
                elif isinstance(base, Decimal):
                    base_decimal = base
                elif isinstance(base, Fraction):  # *
                    base_decimal = _fraction_to_decimal(base)
                elif isinstance(base, float):
                    base_decimal = ctx.create_decimal_from_float(base)
                elif isinstance(base, int):  # ^
                    base_decimal = Decimal(base, context=ctx)
                else:
                    assert_never(base)
                exponent_decimal = _fraction_to_decimal(exponent)
                product_decimal *= ctx.power(base_decimal, exponent_decimal)
            else:
                assert_never(exponent)
        if product_decimal == Decimal(1):
            return product_fraction
        return _fraction_to_decimal(product_fraction) * product_decimal


@dataclass(frozen=True, **slots)
class StdUncertainty:
    """Concise notation for the one-standard-deviation uncertainty of a
    numerical value.

    For example, the parentheses in `12.3456(89) kg` means:
    ```
         numerical value = 12.3456
    standard uncertainty =  0.0089
    ```
    Use `typing.Annotated` to attach the uncertainty information to the value:
    ```py
    from decimal import Decimal
    from typing import Annotated
    from isqx import KG, StdUncertainty

    CONST_FOO: Annotated[Decimal, KG, StdUncertainty(89)] = Decimal("12.3456")
    ```
    """

    # not really used anywhere - its for clarity only.
    # we could try to propagate errors for derived physical constants,
    # but that is overkill.

    value: int
    """The one-standard-deviation uncertainty expressed in the least
    significant digit(s) of the numerical value. Must be greater than zero."""


@final
class _E(SupportsDecimal):
    __slots__ = ()

    def to_decimal(self, ctx: decimal.Context) -> Decimal:
        return ctx.exp(Decimal(1))

    def __float__(self) -> float:
        return math.e

    def __abs__(self) -> _E:
        return self

    def __str__(self) -> str:
        return "𝑒"


@final
class _PI(SupportsDecimal):
    __slots__ = ()

    def to_decimal(self, ctx: decimal.Context) -> Decimal:
        # from: https://docs.python.org/3/library/decimal.html#recipes
        ctx.prec += 2  # extra digits for intermediate steps
        three = Decimal(3)
        lasts, t, s, n, na, d, da = Decimal(0), three, Decimal(3), 1, 0, 0, 24
        while s != lasts:
            lasts = s
            n, na = n + na, na + 8
            d, da = d + da, da + 32
            t = (t * n) / d
            s += t
        ctx.prec -= 2
        pi = +s  # unary plus applies the new precision
        return pi

    def __float__(self) -> float:
        return math.pi

    def __abs__(self) -> _PI:
        return self

    def __str__(self) -> str:
        return "π"


E: Final[_E] = _E()
"""The Euler number [ISO 80000-2:2019 2-13.1][isqx._citations.ISO_80000_2]."""
PI: Final[_PI] = _PI()
"""The ratio of the circumference of a circle to its diameter
[ISO 80000-2:2019 2-14.1][isqx._citations.ISO_80000_2]."""


#
# tools for introspecting `Annotated` (especially module attributes)
# where __doc__ is not available
#


@dataclass(frozen=True, **slots)
class AnnotatedMetadata:
    unit: Expr | None
    std_uncertainty: StdUncertainty | None

    @classmethod
    def from_args(cls, args: Sequence[Any]) -> AnnotatedMetadata:
        """Extract metadata from the args of an `Annotated` type.

        Example:
        ```pycon
        >>> from typing import Annotated
        >>> from isqx import KG, StdUncertainty
        >>> from dataclasses import dataclass
        >>> @dataclass
        ... class Foo:
        ...     class_member: Annotated[float, KG, StdUncertainty(13)]
        ...
        >>> def bar(function_arg: Annotated[float, KG]) -> None:
        ...     ...
        ...
        >>> from isqx import AnnotatedMetadata
        >>> def print_metadata(obj) -> None:
        ...     for name, ann in get_type_hints(obj, include_extras=True).items():
        ...         print(f"{name}: {AnnotatedMetadata.from_args(get_args(ann))}")
        ...
        >>> print_metadata(Foo)
        class_member: AnnotatedMetadata(unit=kilogram, std_uncertainty=StdUncertainty(value=13))
        >>> print_metadata(bar)
        function_arg: AnnotatedMetadata(unit=kilogram)
        return: AnnotatedMetadata(unit=None)
        ```
        """
        unit: _TaggedAllowedExpr | None = None
        std_uncertainty: StdUncertainty | None = None
        for arg in args:
            if isinstance(arg, _ARGS_EXPR):
                if unit is not None:
                    raise DuplicateAnnotationError(args, arg)
                unit = arg
            elif isinstance(arg, StdUncertainty):
                if std_uncertainty is not None:
                    raise DuplicateAnnotationError(args, arg)
                std_uncertainty = arg
        return cls(unit=unit, std_uncertainty=std_uncertainty)

    def __str__(self) -> str:
        kv = []
        for field in fields(self):
            if (value := getattr(self, field.name)) is None:
                continue
            kv.append(f"{field.name}={value}")
        return f"{type(self).__name__}({', '.join(kv)})"


def module_attribute_metadata(
    module_rt: ModuleType,
) -> Generator[tuple[str, AnnotatedMetadata], None, None]:
    """Yields all (name, metadata) pairs for module attributes that are
    annotated with valid [unit expressions][isqx.Expr] and/or
    [standard uncertainties][isqx.StdUncertainty] in the given module.

    Effectively returns the units and uncertainties of all constants. Example:
    ```pycon
    >>> from isqx import module_attribute_metadata, iso80000
    >>> for name, metadata in module_attribute_metadata(iso80000):
    ...     print(f"{name}: {metadata}")
    ...
    CONST_SPEED_OF_LIGHT_VACUUM: AnnotatedMetadata(unit=meter · second⁻¹)
    CONST_PLANCK: AnnotatedMetadata(unit=joule · second
    - joule = newton · meter
    - newton = kilogram · meter · second⁻²)
    CONST_REDUCED_PLANCK: AnnotatedMetadata(unit=joule · second
    - joule = newton · meter
    - newton = kilogram · meter · second⁻²)
    ...
    ```
    """
    for name, anno in get_type_hints(module_rt, include_extras=True).items():
        if get_origin(anno) is not Annotated:
            continue
        metadata = AnnotatedMetadata.from_args(get_args(anno))
        yield name, metadata


@dataclass(frozen=True, **slots)
class Anchor:  # NOTE: not using namedtuple so json serialization works nicely
    """A simple wrapper over a text with an external link.

    It is used to:

    - yield a bare string or a "link" in [`isqx.BasicFormatter`][].
    - create a hyperlink in the [where clause of a definition][isqx.details.WhereFragment]
    - render a link in the [HTML documentation][isqx.mkdocs.extension.MkdocsFormatter].
    """

    text: str
    path: str


StrFragment: TypeAlias = Union[str, Anchor]

#
# errors
#


class IsqxError(Exception):
    """Base exception for all errors raised by the isqx library."""


@dataclass(frozen=True, **slots)
class CompositionError(IsqxError):
    outer: type
    inner: Any
    msg: str
    help: str | None = None

    def __str__(self) -> str:  # pragma: no cover
        outer_name = (
            self.outer.name
            if isinstance(self.outer, Prefix)
            else self.outer.__name__
        )
        inner_repr = f"`{self.inner}` (of type `{type(self.inner).__name__}`)"

        message = (
            f"invalid composition: cannot apply `{outer_name}` to {inner_repr}."
            f"\nreason: {self.msg}"
        )
        if self.help:
            message += f"\n= help: {self.help}"
        return message


@dataclass(frozen=True, **slots)
class MixedKindError(IsqxError):
    terms: tuple[Expr, ...]

    def __str__(self) -> str:  # pragma: no cover
        return (
            "cannot mix expressions of different kinds in a product."
            f"\n  found kinds: {', '.join(f'`{kind(t)}`' for t in self.terms)}"
            "\n= help: all terms in a `Mul` expression must be of the same kind "
            "(e.g., all units like `M` and `S`, or all dimensions like "
            "`DIM_LENGTH` and `DIM_TIME`)."
        )


# conversion
@dataclass(frozen=True, **slots)
class KindMismatchError(IsqxError):
    origin_kind: ExprKind
    target_kind: ExprKind

    def __str__(self) -> str:  # pragma: no cover
        return (
            "cannot convert between expressions of different kinds."
            f"\nnote: origin kind: `{self.origin_kind}`"
            f"\n      target kind: `{self.target_kind}`"
        )


@dataclass(frozen=True, **slots)
class DimensionMismatchError(IsqxError):
    origin: Expr
    target: Expr
    dim_origin: Expr
    dim_target: Expr

    def __str__(self) -> str:  # pragma: no cover
        return (
            f"cannot convert from `{self.origin}` to `{self.target}`."
            "\n= help: expected compatible dimensions, but found:"
            f"\ndimension of origin: `{self.dim_origin}`"
            f"\ndimension of target: `{self.dim_target}`"
        )


@dataclass(frozen=True, **slots)
class UnitKindMismatchError(IsqxError):
    qtykind: QtyKind
    unit: Expr
    dim_kind: Expr
    dim_unit: Expr

    def __str__(self) -> str:  # pragma: no cover
        return (
            f"cannot create tagged unit for kind `{self.qtykind.tags}` with "
            f"unit `{self.unit}`."
            f"\nexpected dimension of kind: `{self.dim_kind}`"
            f" (`{self.qtykind.unit_si_coherent}`)"
            f"\n   found dimension of unit: `{self.dim_unit}`"
            f" (`{self.unit}`)"
        )


@dataclass(frozen=True, **slots)
class NonLinearConversionError(IsqxError):
    origin: Expr
    target: Expr

    def __str__(self) -> str:  # pragma: no cover
        return (
            f"cannot create a value-agnostic converter from `{self.origin}` "
            f"to `{self.target}`."
            "conversion between a logarithmic and a linear unit is non-linear."
            "\n= help: this requires a reference value (e.g., 1V for dBV), "
            "but the `isqx` library only performs value-agnostic conversions. "
            "perform the calculation manually instead."
        )


@dataclass(frozen=True, **slots)
class DuplicateAnnotationError(IsqxError):
    annotation_args: Sequence[Any]
    conflicting_arg: _TaggedAllowedExpr | StdUncertainty

    def __str__(self) -> str:  # pragma: no cover
        return (
            f"duplicate annotation `{self.conflicting_arg}` found in "
            f"{', '.join(repr(arg) for arg in self.annotation_args)}."
            "\n= help: only one unit and one standard uncertainty can be "
            "applied."
        )



"""[Dimensionless quantities][isqx.Dimensionless] and
[quantity kinds][isqx.QtyKind] themselves do not store the precise definition
of how it is derived.

This module contains **optional** details that "link" them together. It is
stored in a [dictionary of expressions to its details][isqx.details.Details]. A
detail can contain multiple [equations][isqx.details.Equation] or [common
symbols][isqx.details.Symbol].

Our [mkdocs plugin][isqx.mkdocs.plugin] reads dictionaries from this module to
"inject" information into docstrings.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Protocol, Union, runtime_checkable

from typing_extensions import TypeAlias, get_args

from .._core import (
    Dimensionless,
    LazyProduct,
    Number,
    QtyKind,
    Scaled,
    StrFragment,
    Tagged,
    slots,
)


@dataclass(frozen=True)
class _RefSelf:
    __slots__ = ()


SELF = _RefSelf()
"""A special marker to represent a reference to "itself" in the meaning of a
[definition's where clause][isqx.details.Equation.where]."""

RefDetail: TypeAlias = Union[
    QtyKind, Dimensionless, Tagged, LazyProduct, Number
]
"""A reference to a quantity kind, dimensionless number, a numerical constant,
or a string"""
WhereFragment: TypeAlias = Union[_RefSelf, RefDetail, StrFragment]
# allow factories: Callable[..., RefDetail]
WhereValue: TypeAlias = Union[WhereFragment, tuple[WhereFragment, ...]]
Where: TypeAlias = dict[str, WhereValue]
# allow the key of the where to allow a symbol, e.g. m(X) where X is a substance
# but this will be very complicated.


@dataclass(frozen=True, **slots)
class Wikidata:
    """Stores the Wikidata Q-code for a quantity kind."""

    qcode: str


# isolating into a protocol to help with static analysis in griffe
@runtime_checkable
class HasKaTeXWhere(Protocol):
    katex: str
    r"""A $\KaTeX$ string"""
    where: Where | None
    r"""A mapping of the symbols in the $\KaTeX$ string to their meanings."""


@dataclass(frozen=True, **slots)
class Symbol(HasKaTeXWhere):
    """Stores a common symbol for a quantity kind.

    Since the [equation][isqx.details.Equation] class already contains the
    symbol, adding the symbol is not necessary. However, it is useful when the
    quantity kind is not defined by an equation.
    """

    katex: str
    where: Where | None = None
    remarks: str | None = None
    """A description of when this symbol should be used, e.g. for mole fraction,
    `x_X` is used for condensed phase and `y_X` is used for gaseous mixtures."""


@dataclass(frozen=True, **slots)
class Equation(HasKaTeXWhere):
    """Stores the equation for a quantity kind."""

    katex: str
    where: Where | None = None
    assumptions: set[StrFragment | tuple[StrFragment, ...]] | None = None
    """A set of assumptions under which the definition is valid."""


DetailKey: TypeAlias = Union[
    QtyKind, Dimensionless, Tagged, Scaled, LazyProduct, Number
]
# support adding symbols to tagged(dimensionless) and -1 * Log...
_ARGS_DETAIL_KEY = get_args(DetailKey)
Detail: TypeAlias = Union[Wikidata, Symbol, Equation]
Details: TypeAlias = dict[
    Union[DetailKey, Callable[..., DetailKey]],
    Union[Detail, tuple[Detail, ...]],
]



#!/usr/bin/env -S uv run --script
import logging
import shutil
import subprocess
from functools import partial
from pathlib import Path

PATH_ROOT = Path(__file__).parent.parent

FILENAME_INDEX = "index"
FILENAME_CONTRIBUTING = "contributing"
logger = logging.getLogger(__name__)


def copy_docs_file(path_in: Path, path_out: Path) -> None:
    """Copy a file from `path_in` to `path_out` with warnings.

    Should be used to sync the homepage in the docs with the one from
    `path_root/README.md`.

    - `base_path: !relative $config_dir` and `--8<-- "README.md"` doesn't work
      and it causes crossrefs to break.
    - symlinking `docs/README.md` to `../README.md` breaks images as well.
    - https://stackoverflow.com/questions/75716969/mkdocs-with-readme-as-index-containing-images-with-broken-links
    - https://stackoverflow.com/questions/73828765/mkdocs-how-to-link-to-the-same-images-in-readme-md-and-docs-index-md-at-the
    """
    with open(path_in, "r") as f:
        readme_content = f.read()
    output = (
        "<!-- DO NOT EDIT! CHANGES WILL BE LOST. edit `README.md` and"
        + f" run `{Path(__file__).relative_to(PATH_ROOT)}` instead. -->\n"
        + readme_content.replace("](docs/assets/", "](assets/")
    )
    with open(path_out, "w+") as f:
        content = f.read()
        if content != output:
            f.write(output)


copy_readme = partial(
    copy_docs_file,
    path_in=PATH_ROOT / "README.md",
    path_out=PATH_ROOT / "docs" / f"{FILENAME_INDEX}.md",
)

copy_contributing = partial(
    copy_docs_file,
    path_in=PATH_ROOT / "CONTRIBUTING.md",
    path_out=PATH_ROOT / "docs" / f"{FILENAME_CONTRIBUTING}.md",
)


# mkdocs events


def on_pre_build(config) -> None:  # type: ignore
    copy_readme()
    copy_contributing()


PATH_VIS_SOURCE = Path(__file__).parent.parent / "src" / "isqx_vis"


def on_post_build(config) -> None:  # type: ignore
    path_out = Path(config["site_dir"])
    path_vis_html = path_out / "vis.html"

    if not (PATH_VIS_SOURCE / "node_modules").exists():
        logger.error(
            "`module_modules` not found.\n"
            f"= help: run `pnpm install` in {PATH_VIS_SOURCE}"
        )
        return
    subprocess.run(["pnpm", "build"], cwd=PATH_VIS_SOURCE, check=True)
    # not removing dir because assets/objects.json exists
    shutil.copytree(
        PATH_VIS_SOURCE / "dist",
        path_out,
        ignore=lambda d, files: ["index.html"] if "index.html" in files else [],
        dirs_exist_ok=True,
    )
    shutil.copy(
        PATH_VIS_SOURCE / "dist" / "index.html",
        path_vis_html,
    )
    logger.info(f"built isqx-vis to {path_vis_html}")


if __name__ == "__main__":
    copy_readme()
    copy_contributing()



# mypy: ignore-errors
# %%
# --8<-- [start:tsfc_example]
import isqx
import isqx.usc as usc

tsfc_converter = isqx.convert(
    usc.LB / (isqx.HOUR * usc.LBF), isqx.KG / (isqx.S * isqx.N), exact=True
)
print(f"1 lb/(h·lbf) = {tsfc_converter.scale} g/(s·kN)")
# --8<-- [end:tsfc_example]
# %%
"""
# --8<-- [start:tsfc_output]
1 lb/(h·lbf) = 50/1765197 g/(s·kN)
# --8<-- [end:tsfc_output]
"""
# %%
# --8<-- [start:verbose_fmt_example]
print(isqx.SPECIFIC_HEAT_CAPACITY(usc.BTU_IT * usc.LB**-1 * usc.R**-1))
# --8<-- [end:verbose_fmt_example]
# %%
"""
# --8<-- [start:verbose_fmt_output]
(btu_it · pound⁻¹ · rankine⁻¹)['specific_heat_capacity']
- btu_it = 1055.05585262 · joule
  - joule = newton · meter
    - newton = kilogram · meter · second⁻²
- pound = 0.45359237 · kilogram
- rankine = 5/9 · kelvin
# --8<-- [end:verbose_fmt_output]
"""
# %%
# --8<-- [start:tagged_error_example]
import isqx.aerospace as aero

try:
    isqx.convert(
        aero.PRESSURE_ALTITUDE(isqx.M), aero.GEOMETRIC_ALTITUDE(isqx.M)
    )
except isqx.DimensionMismatchError as e:
    print(e)
# --8<-- [end:tagged_error_example]
# %%
"""
# --8<-- [start:tagged_error_output]
cannot convert from `meter['altitude', relative to `'mean_sea_level'`, 'pressure']` to `meter['altitude', relative to `'mean_sea_level'`, 'geometric']`.
= help: expected compatible dimensions, but found:
dimension of origin: `L['altitude', relative to `'mean_sea_level'`, 'pressure']`
dimension of target: `L['altitude', relative to `'mean_sea_level'`, 'geometric']`
# --8<-- [end:tagged_error_output]
"""
# %%
# --8<-- [start:decorator_def]
import typing as t
from dataclasses import dataclass

from typing_extensions import ParamSpec

T = t.TypeVar("T")
P = ParamSpec("P")


@t.overload
def sphinx_doc(
    *, formatter: t.Union[isqx.Formatter, str]
) -> t.Callable[[T], T]: ...


@t.overload
def sphinx_doc(obj: T, /) -> T: ...


def sphinx_doc(
    obj: t.Optional[T] = None,
    /,
    *,
    formatter: t.Union[isqx.Formatter, str] = isqx.BasicFormatter(
        verbose=False
    ),
):
    def wrapper(obj_: T) -> T:
        doc = obj_.__doc__ or ""
        if doc.startswith(obj_.__qualname__ + "("):
            doc = ""

        hints = t.get_type_hints(obj_, include_extras=True)
        params_doc = []
        for name, hint in hints.items():
            if t.get_origin(hint) is not t.Annotated:
                continue
            # assuming the first metadata item is the isqx unit expression
            unit_expr = t.get_args(hint)[1]
            unit_str = isqx.fmt(unit_expr, formatter=formatter)
            directive = ":returns:" if name == "return" else f":param {name}:"
            params_doc.append(f"{directive} `{unit_str}`")

        if doc and params_doc:
            doc += "\n\n" + "\n".join(params_doc)
        elif params_doc:
            doc = "\n".join(params_doc)

        obj_.__doc__ = doc.strip()
        return obj_

    if obj is None:
        return wrapper
    return wrapper(obj)


# --8<-- [end:decorator_def]
# %%
# --8<-- [start:breguet_example]
_T = t.TypeVar("_T")

MPerSCruiseTas = t.Annotated[_T, aero.TRUE_AIRSPEED["cruise"](isqx.M_PERS)]
LOverD = t.Annotated[_T, isqx.ratio(isqx.LIFT(isqx.N), isqx.DRAG(isqx.N))]
SIsp = t.Annotated[_T, isqx.TIME["specific_impulse"](isqx.S)]
WiOverWf = t.Annotated[
    _T, isqx.ratio(aero.TAKEOFF_MASS(isqx.KG), aero.LANDING_MASS(isqx.KG))
]
Range = t.Annotated[_T, isqx.DISTANCE(isqx.M)]


@sphinx_doc
def breguet_range(
    v: MPerSCruiseTas,
    l_over_d: LOverD,
    isp: SIsp,
    wi_over_wf: WiOverWf,
    *,
    xp,  # library that implements the array api
) -> Range:
    """Calculates the Breguet range of an aircraft."""
    return v * l_over_d * isp * xp.log(wi_over_wf)


print(breguet_range.__doc__)
# --8<-- [end:breguet_example]
# %%
"""
# --8<-- [start:breguet_output]
Calculates the Breguet range of an aircraft.

:param v: `(meter · second⁻¹)['airspeed', 'true', 'cruise']`
:param l_over_d: `ratio[`newton[_Tensor(rank=1), 'lift']` to `newton[_Tensor(rank=1), 'drag']`]`
:param isp: `second['specific_impulse']`
:param wi_over_wf: `ratio[`kilogram['aircraft', 'takeoff']` to `kilogram['aircraft', 'landing']`]`
:returns: `meter['distance']`
# --8<-- [end:breguet_output]
"""


# %%
# --8<-- [start:dataclass_example]
@sphinx_doc
@dataclass
class AircraftParams:
    """Aircraft performance parameters."""

    l_over_d: LOverD
    specific_impulse: SIsp


print(AircraftParams.__doc__)
# --8<-- [end:dataclass_example]
"""
# --8<-- [start:dataclass_output]
Aircraft performance parameters.

:param l_over_d: `ratio[`newton[_Tensor(rank=1), 'lift']` to `newton[_Tensor(rank=1), 'drag']`]`
:param specific_impulse: `second['specific_impulse']`
# --8<-- [end:dataclass_output]
"""
# %%



from fractions import Fraction

from isqx import CENTI, HOUR, G, M, S, simplify
from isqx._fmt import BasicFormatter, fmt
from isqx.usc import BTU_IT, FAHRENHEIT, FT, IN, R

K_VALUE = BTU_IT * IN / (HOUR * FT**2 * R)
CM = CENTI * M
DYN = (G * CM * S**-2).alias("dyn")
STATC = (DYN ** Fraction(1, 2) * M).alias("statc")


def test_fmt_basic_tagged() -> None:
    from isqx import M_PERS
    from isqx.aerospace import TRUE_AIRSPEED

    M_PERS_TAS = TRUE_AIRSPEED(M_PERS)
    assert (
        fmt(M_PERS_TAS, formatter=BasicFormatter())
        == "(meter · second⁻¹)['airspeed', 'true']"
    )


def test_fmt_basic_translated() -> None:
    assert (
        fmt(FAHRENHEIT, formatter=BasicFormatter(verbose=True))
        == """fahrenheit
- fahrenheit = rankine - 459.67
  - rankine = 5/9 · kelvin"""
    )


def test_fmt_basic_log() -> None:
    from isqx import DB

    assert (
        fmt(DB, formatter=BasicFormatter(verbose=True))
        == """decibel
- bel = log₁₀(ratio)"""
    )


def test_fmt_basic_log_level() -> None:
    from isqx import DBU

    assert (
        fmt(DBU, formatter=BasicFormatter(verbose=True))
        == """dBu
- dBu = 20 · log₁₀(ratio[`volt` to `0.6¹⸍² · volt`])
  - volt = watt · ampere⁻¹
    - watt = joule · second⁻¹
      - joule = newton · meter
        - newton = kilogram · meter · second⁻²"""
    )


def test_fmt_basic_k_value() -> None:
    assert (
        fmt(K_VALUE, formatter=BasicFormatter(verbose=True))
        == """btu_it · inch · (hour · foot² · rankine)⁻¹
- btu_it = 1055.05585262 · joule
  - joule = newton · meter
    - newton = kilogram · meter · second⁻²
- inch = 1/12 · foot
  - foot = 0.3048 · meter
- hour = 60 · minute
  - minute = 60 · second
- rankine = 5/9 · kelvin"""
    )
    assert (
        fmt(simplify(K_VALUE), formatter=BasicFormatter(verbose=True))
        == "1055.05585262 · 1/12 · 0.3048 · 60⁻¹ · 60⁻¹ · 0.3048⁻² · (5/9)⁻¹ · (meter · kilogram · second⁻³ · kelvin⁻¹)"
    )


def test_fmt_basic_statc() -> None:
    assert (
        fmt(STATC, formatter=BasicFormatter(verbose=True))
        == """statc
- statc = dyn¹⸍² · meter
  - dyn = gram · centimeter · second⁻²
    - gram = 1/1000 · kilogram"""
    )
    assert (
        fmt(simplify(STATC), formatter=BasicFormatter(verbose=True))
        == "(1/1000)¹⸍² · (1/100)¹⸍² · (meter³⸍² · kilogram¹⸍² · second⁻¹)"
    )



# mypy: ignore-errors
from pint import UnitRegistry

ureg = UnitRegistry()
RHO_0 = 1.225 * (ureg.kilogram * ureg.meter**-3)
"""Density of air at mean sea level"""


def eas(tas, air_density):
    """Converts TAS to EAS.

    :param tas: true airspeed (meters per second)
    :param air_density: atmospheric density (kilograms per meter cubed)
    :returns: equivalent airspeed (meters per second)
    """
    return tas * (air_density / RHO_0) ** 0.5


RHO_0_PURE_PY = 1.225


def eas_pure_py(tas, air_density):
    return tas * (air_density / RHO_0_PURE_PY) ** 0.5



"""A `griffe` extension to dynamically collect `isqx` definitions/details.

[Details][isqx.details.Details] are injected into the docstrings of attributes
and functions in the documentation generated by `mkdocstrings-python`.

!!! note

    You must install `isqx` with the `docs` extra optional dependencies
    to use this module.
"""

from __future__ import annotations

import ast
import inspect
import json
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, NamedTuple, Union

from griffe import (
    Alias,
    Docstring,
    Extension,
    Inspector,
    Object,
    ObjectNode,
    Visitor,
    dynamic_import,
    get_logger,
)
from griffe._internal.expressions import (
    Expr,
    ExprAttribute,
    ExprCall,
    ExprDict,
    ExprKeyword,
    ExprName,
    ExprSubscript,
    ExprTuple,
)
from griffe._internal.models import Attribute
from typing_extensions import TypeAlias

from .. import (
    Anchor,
    BasicFormatter,
    Dimensionless,
    LazyProduct,
    QtyKind,
    StrFragment,
    Tagged,
    kind,
    module_attribute_metadata,
)
from .. import (
    Expr as IsqxExpr,
)
from .._core import _ARGS_EXPR, _ARGS_NUMBER
from ..details import (
    _ARGS_DETAIL_KEY,
    Equation,
    HasKaTeXWhere,
    Symbol,
    Wikidata,
    _RefSelf,
)
from . import PATH_PLUGIN

if TYPE_CHECKING:
    from typing import Generator

    from griffe import GriffeLoader, Module
    from griffe._internal.models import Attribute

    from .. import AnnotatedMetadata, NamedExpr, Number
    from .._fmt import _BasicFormatterState
    from ..details import Detail, DetailKey, Details, RefDetail, WhereValue
    from .plugin import IsqxPluginConfig


logger = get_logger(__name__)


def get_templates_path() -> Path:
    """See: https://mkdocstrings.github.io/usage/handlers/#handler-extensions"""
    return PATH_PLUGIN / "templates"


_ARGS_DEFINITION = (LazyProduct, *_ARGS_NUMBER, *_ARGS_EXPR, QtyKind)
# NOTE: equilibrium constants are factory functions that return a QtyKind.
# right now we skip over them!

Unit: TypeAlias = Union[tuple[StrFragment, ...], None]
"""The unit of measurement *rendered* as string fragments."""
# Mul(M, S) is represented as (Anchor("M", "isqx.M"), "\cdot", Anchor("S", "isqx.S"))
# in the future, we want to return a tagged union (the direct representation
# of the unit). for now, this is good enough.
# we also want to get rid of the None.


# see: https://mkdocstrings.github.io/griffe/guide/users/how-to/selectively-inspect/
@dataclass(frozen=True)
class Where:
    symbol: str
    description: StrFragment | tuple[StrFragment, ...]
    unit: Unit  # for now


@dataclass(frozen=True)
class KaTeXWhere:
    katex: str
    where: tuple[Where, ...] | None = None


@dataclass(frozen=True)
class SymbolDetail(KaTeXWhere):
    remarks: str | None = None


@dataclass(frozen=True)
class EquationDetail(KaTeXWhere):
    assumptions: tuple[StrFragment | tuple[StrFragment, ...], ...] | None = None


@dataclass(frozen=True)
class WikidataDetail:
    qcode: str


@dataclass(frozen=True)
class Quantity:
    value: str  # for now: in the future we should support lazy products
    unit: Unit  # for now


CanonicalPath: TypeAlias = str


@dataclass
class QtyKindDetail:
    """Several quantities can share the same unit. Two cases:

    - Explicit inheritance: when a quantity is defined by subscripting
        another, like `POTENTIAL_ENERGY = ENERGY["potential"]`: this is a
        strong signal of a parent-child relationship.
        we can detect this during the griffe AST walking
    - Implicit grouping: some quantities are related by their physical
        dimension but not defined via subscripting (e.g., SPEED and VELOCITY).
        we can group them by their underlying SI dimension.
        the hierarchy can be inferred from the tags.
    """

    parent: str | None = None
    unit_si_coherent: Unit = None  # str for now
    tags: tuple[str, ...] = field(default_factory=tuple)  # str for now
    wikidata: list[WikidataDetail] = field(default_factory=list)
    symbols: list[SymbolDetail] = field(default_factory=list)
    equations: list[EquationDetail] = field(default_factory=list)


class IsqxExtension(Extension):
    def __init__(
        self, config: IsqxPluginConfig, objects_out_path: str | None = None
    ):
        self.config = config
        self.definitions: Definitions = {}
        self.objects: dict[str, QtyKindDetail] = {}  # by path
        self.objects_out_path = objects_out_path
        self.possible_parent_maps: dict[str, str] = {}  # child -> parent
        """Note that items in this map are not guaranteed to be isqx objects
        because griffe works with static analysis."""

    def on_instance(
        self,
        *,
        node: ast.AST | ObjectNode,
        obj: Object,
        agent: Visitor | Inspector,
        **kwargs: Any,
    ) -> None:
        inject_citation_into_docstring(obj, agent)

        # collect strict parent-child relationships
        if (
            isinstance(obj, Attribute)
            and isinstance(obj.value, ExprSubscript)
            and isinstance(
                parent_expr := obj.value.left, (ExprName, ExprAttribute)
            )
        ):
            parent_path = parent_expr.canonical_path
            if parent_path.split(".", 1)[0] in ("typing", "typing_extensions"):
                # heuristic: we captured something invalid like Union[str, int]
                return
            self.possible_parent_maps[obj.canonical_path] = parent_path

    def on_module_instance(
        self,
        *,
        node: ast.AST | ObjectNode,
        mod: Module,
        agent: Visitor | Inspector,
        **kwargs: Any,
    ) -> None:
        """Populate definitions from the current module being visited using
        runtime analysis, if it is matches the path under the configured
        [definitions][isqx.mkdocs.plugin.IsqxPluginConfig.definitions].

        Note that if module `a` contains `FOO: Annotated[float, isqx.M] = 1.0`
        and module `b` contains `from a import FOO`, then `self.definitions`
        will contain both (which points to the same object):

        - `a.FOO`: Definition(value=1.0, annotated_metadata=M)
        - `b.FOO`: Definition(value=1.0, annotated_metadata=None)

        which is not ideal. When `griffe` visits `b`, `on_alias` will make sure
        the definition for `b.FOO` is removed.
        """
        if mod.path not in self.config.definitions:
            return
        module_rt = dynamic_import(mod.path)
        metadata = dict(module_attribute_metadata(module_rt))
        self.definitions |= {
            f"{mod.path}.{name}": Definition(obj, metadata.get(name))
            for name, obj in inspect.getmembers(module_rt)
            if isinstance(obj, _ARGS_DEFINITION)
        }

    def on_alias_instance(
        self,
        *,
        node: ast.AST | ObjectNode,
        alias: Alias,
        agent: Visitor | Inspector,
        **kwargs: Any,
    ) -> None:
        # todo: actually we probably shouldn't do this
        if alias.path in self.definitions:
            del self.definitions[alias.path]

    def on_package(
        self, *, pkg: Module, loader: GriffeLoader, **kwargs: Any
    ) -> None:
        logger.info(f"loaded {len(self.definitions)} definitions")
        reverse_definitions = _build_reverse_definitions(self.definitions)
        formatter = MkdocsFormatter(reverse_definitions)

        for defs_path in self.config.details:
            for path, item in _process_details_module(
                defs_path, loader, formatter, reverse_definitions
            ):
                objects = self.objects.setdefault(path, QtyKindDetail())
                attr_target: Attribute = loader.modules_collection[path]
                isqx_extras = _get_or_init_isqx_extras(attr_target)

                if isinstance(item, WikidataDetail):
                    objects.wikidata.append(item)
                    isqx_extras["wikidata"].append(item)
                elif isinstance(item, SymbolDetail):
                    objects.symbols.append(item)
                    isqx_extras["symbols"].append(item)
                elif isinstance(item, EquationDetail):
                    objects.equations.append(item)
                    isqx_extras["details"].append(item)

        # set the parent (if found). would be nice for jinja to also show this
        # but we focus on the frontend vis for now.
        for child_path, parent_path in self.possible_parent_maps.items():
            if child_path in self.objects and parent_path in self.objects:
                child_object_detail = self.objects[child_path]
                child_object_detail.parent = parent_path

        for path, details in self.objects.items():
            if path not in self.definitions:
                continue
            definition = self.definitions[path]
            if not isinstance(v := definition.value, QtyKind):
                continue
            details.unit_si_coherent = tuple(
                formatter.fmt(v.unit_si_coherent)
            )  # str for now
            if v.tags is not None:
                details.tags = tuple(str(t) for t in v.tags)  # str for now

        if self.objects_out_path:
            constants = {}
            for path, definition in self.definitions.items():
                if not isinstance(
                    definition.value, (LazyProduct, *_ARGS_NUMBER)
                ):
                    continue
                value = definition.value
                value_str = str(
                    value.to_approx()
                    if isinstance(value, LazyProduct)
                    else value
                )
                unit_expr: IsqxExpr | None = None
                if not (
                    (meta := definition.annotated_metadata)
                    and (unit_expr := meta.unit)
                ):
                    logger.warning(
                        f"no unit found for {path} in annotated metadata"
                    )
                    continue  # dimensionless quantities should be annotated anyways
                constants[path] = Quantity(
                    value=value_str,
                    unit=(
                        tuple(formatter.fmt(unit_expr))  # str for now
                        if unit_expr is not None
                        else None
                    ),
                )

            json_path = Path(self.objects_out_path) / "objects.json"
            # do not serialize members that have `None` or empty tuple values to
            # save space
            output_data = {
                "qtyKinds": {
                    path: to_dict(details)
                    for path, details in self.objects.items()
                },
                "constants": {
                    path: to_dict(details)
                    for path, details in constants.items()
                },
                "units": {},  # TODO
            }
            json_path.parent.mkdir(parents=True, exist_ok=True)
            with open(json_path, "w") as f:
                json.dump(output_data, f)
            logger.info(f"wrote objects to {json_path}")


def to_dict(obj: Any) -> dict[str, Any]:
    return asdict(obj, dict_factory=lambda x: {k: v for (k, v) in x if v})


class Definition(NamedTuple):
    value: LazyProduct | Number | IsqxExpr | QtyKind
    annotated_metadata: AnnotatedMetadata | None


Definitions: TypeAlias = dict[CanonicalPath, Definition]


class _ReverseDefinition(NamedTuple):
    path: CanonicalPath
    annotated_metadata: AnnotatedMetadata | None


_ReverseDefinitions: TypeAlias = dict[int, _ReverseDefinition]
"""A mapping of the `builtins.id(value)` to the reverse definition."""
#
# unit formatting
#


class MkdocsFormatter(BasicFormatter):
    def __init__(
        self,
        reverse_definitions: _ReverseDefinitions,
    ):
        self.reverse_definitions = reverse_definitions
        super().__init__(verbose=False)

    def visit_named(
        self, expr: NamedExpr, state: _BasicFormatterState
    ) -> Generator[StrFragment, None, None]:
        name_formatted: str = next(super().visit_named(expr, state))  # type: ignore
        if reverse_def := self.reverse_definitions.get(id(expr)):
            yield Anchor(
                text=name_formatted,
                path=reverse_def.path,
            )
        else:
            yield name_formatted


#
# details
#


def parse_katex_where_static(
    katex_where_rt: HasKaTeXWhere,
    katex_where_st: str | Expr,
) -> tuple[str, ExprDict | None]:
    # working with static analysis is very annoying, we have to handle kwargs
    # and defaults. using a bunch of asserts for now.
    assert isinstance(katex_where_st, ExprCall), katex_where_st
    katex_st = katex_where_st.arguments[0]
    katex_st_value = (
        katex_st.value if isinstance(katex_st, ExprKeyword) else katex_st
    )
    assert isinstance(katex_st_value, str), katex_st_value
    katex_st_value = (
        katex_st_value.removeprefix("'")
        .removeprefix('"')
        .removesuffix("'")
        .removesuffix('"')
        .replace("\\\\", "\\")
    )  # regex is probably better but eh
    if not len(katex_where_st.arguments) > 1:
        return katex_st_value, None
    where_st = katex_where_st.arguments[1]
    where_st_value = where_st
    if isinstance(where_st, ExprKeyword):
        if where_st.name != "where":
            return katex_st_value, None
        where_st_value = where_st.value
    assert isinstance(where_st_value, ExprDict), where_st_value
    assert katex_where_rt.katex == katex_st_value, (
        f"`griffe` skipped an entry!"
        "\nthis can happen in two keys in the dictionary refer to the same "
        "value. check the definitions."
        f"\nruntime value {katex_where_rt.katex}\ndoes not match griffe: {katex_st_value}"
    )
    return katex_st_value, where_st_value


def parse_where(
    key_rt: DetailKey | Callable[..., DetailKey],
    detail_rt: HasKaTeXWhere,
    where_st: ExprDict,
    *,
    self_path: str,
    reverse_definitions: _ReverseDefinitions,
    formatter: MkdocsFormatter,
    mut_self_symbols: list[str],
) -> Generator[Where, None, None]:
    if detail_rt.where is None:
        return
    for (w_rt_k, w_rt_v), w_st_v in zip(
        detail_rt.where.items(),
        where_st.values,
    ):
        description = _parse_fragments(w_rt_v, w_st_v, self_path)
        unit_expr: IsqxExpr | None = None
        frags_for_unit_inference = (
            w_rt_v if isinstance(w_rt_v, tuple) else (w_rt_v,)
        )
        for frag in frags_for_unit_inference:
            if isinstance(frag, (str, Anchor)):
                continue
            if isinstance(frag, _RefSelf):
                mut_self_symbols.append(w_rt_k)
                if isinstance(key_rt, _ARGS_DETAIL_KEY):
                    unit_expr = _get_unit_expr(key_rt, reverse_definitions)
            else:
                unit_expr = _get_unit_expr(frag, reverse_definitions)
            if unit_expr is not None:
                break  # only infer unit from the first fragment

        unit: Unit
        if unit_expr is None:
            unit = None
        elif kind(unit_expr) == "dimensionless":
            unit = ("dimensionless",)
        else:
            unit = tuple(s for s in formatter.fmt(unit_expr))  # str  for now

        yield Where(
            symbol=w_rt_k,
            description=description,
            unit=unit,
        )


def _build_reverse_definitions(definitions: Definitions) -> _ReverseDefinitions:
    reverse_definitions: _ReverseDefinitions = {}
    for path, definition in definitions.items():
        i = id(definition.value)
        if i in reverse_definitions:
            logger.warning(
                f"duplicate definition id for {path} and {reverse_definitions[i].path}"
            )
            continue
        reverse_definitions[i] = _ReverseDefinition(
            path, definition.annotated_metadata
        )
    return reverse_definitions


def _get_or_init_isqx_extras(attr_target: Attribute) -> dict[str, Any]:
    mkdocstrings = attr_target.extra.setdefault("mkdocstrings", {})
    tmpl = "function" if attr_target.is_function else "attribute"
    mkdocstrings.setdefault("template", f"isqx/{tmpl}.html.jinja")
    isqx: dict[str, Any] = mkdocstrings.setdefault("isqx", {})
    isqx.setdefault("details", [])
    isqx.setdefault("symbols", [])
    isqx.setdefault("wikidata", [])
    return isqx


def _process_katex_where_detail(
    detail_rt: HasKaTeXWhere,
    detail_st: str | Expr,
    key_rt: DetailKey | Callable[..., DetailKey],
    self_path: str,
    formatter: MkdocsFormatter,
    reverse_definitions: _ReverseDefinitions,
) -> tuple[SymbolDetail | EquationDetail, list[str]]:
    _katex_st_value, where_st_value = parse_katex_where_static(
        detail_rt, detail_st
    )
    self_symbols: list[str] = []
    where = (
        tuple(
            parse_where(
                key_rt,
                detail_rt,
                where_st_value,
                self_path=self_path,
                formatter=formatter,
                reverse_definitions=reverse_definitions,
                mut_self_symbols=self_symbols,
            )
        )
        if where_st_value is not None
        else None
    )
    if isinstance(detail_rt, Equation):
        return EquationDetail(
            katex=detail_rt.katex,
            where=where,
            assumptions=(
                tuple(a) if (a := detail_rt.assumptions) is not None else None
            ),
        ), self_symbols
    elif isinstance(detail_rt, Symbol):
        return SymbolDetail(
            katex=detail_rt.katex,
            where=where,
            remarks=detail_rt.remarks,
        ), self_symbols
    raise ValueError(f"unknown {type(detail_rt)=}")


def _process_definitions(
    defs_rt: Detail | tuple[Detail, ...],
    defs_st: Expr | str,
    key_rt: DetailKey | Callable[..., DetailKey],
    self_path: str,
    formatter: MkdocsFormatter,
    reverse_definitions: _ReverseDefinitions,
) -> Generator[WikidataDetail | SymbolDetail | EquationDetail, None, None]:
    definitions_rt = defs_rt if isinstance(defs_rt, tuple) else (defs_rt,)
    definitions_st = (
        defs_st.elements if isinstance(defs_st, ExprTuple) else (defs_st,)
    )

    for detail_rt, detail_st in zip(definitions_rt, definitions_st):
        if isinstance(detail_rt, Wikidata):
            yield WikidataDetail(qcode=detail_rt.qcode)
        elif isinstance(detail_rt, HasKaTeXWhere):
            item, self_symbols = _process_katex_where_detail(
                detail_rt,
                detail_st,
                key_rt,
                self_path,
                formatter,
                reverse_definitions,
            )
            yield item
            if self_symbols:
                if len(self_symbols) > 1:
                    raise ValueError(
                        f"too many self symbols for {detail_rt.katex=}"
                    )
                yield SymbolDetail(katex=self_symbols[0])
        else:
            raise ValueError(f"unknown {type(detail_rt)=}")


def _process_details_module(
    defs_path: str,
    loader: GriffeLoader,
    formatter: MkdocsFormatter,
    reverse_definitions: _ReverseDefinitions,
) -> Generator[
    tuple[str, WikidataDetail | SymbolDetail | EquationDetail], None, None
]:
    details_rt: Details = dynamic_import(defs_path)
    details_st: Attribute = loader.modules_collection[defs_path]
    assert isinstance(details_st.value, ExprDict), details_st

    for (key_st, defs_st), (key_rt, defs_rt) in zip(
        zip(details_st.value.keys, details_st.value.values),
        details_rt.items(),
    ):
        assert isinstance(key_st, (ExprAttribute, ExprName)), key_st
        self_path = key_st.canonical_path
        for item in _process_definitions(
            defs_rt,
            defs_st,
            key_rt,
            self_path,
            formatter,
            reverse_definitions,
        ):
            yield (self_path, item)


def _parse_fragments(
    where_rt: WhereValue, where_st: Expr | str, self_path: str
) -> StrFragment | tuple[StrFragment, ...]:
    if isinstance(where_rt, tuple):
        fragments_rt = where_rt
        fragments_st = where_st.elements  # type: ignore
        parsed_frags = tuple(
            _parse_fragment(rt, st, self_path, i)
            for i, (rt, st) in enumerate(zip(fragments_rt, fragments_st))
        )
        return parsed_frags
    return _parse_fragment(where_rt, where_st, self_path, 0)


def _parse_fragment(
    frag_rt: WhereValue, frag_st: Expr | str, self_path: str, i: int
) -> StrFragment:
    if isinstance(frag_rt, (str, Anchor)):
        return frag_rt
    if isinstance(frag_rt, _RefSelf):
        return Anchor(screaming_to_normal(self_path, i), self_path)
    # RefDetail
    assert isinstance(frag_st, (ExprAttribute, ExprName)), (
        "expected value of the where clause to be a reference to a pre-defined "
        f"quantity, but got {frag_st} (an inline expression)\n"
        "= help: define the quantity first and then use it in the where clause"
    )
    path = frag_st.canonical_path
    return Anchor(screaming_to_normal(path, i), path)


def _get_unit_expr(
    meaning: RefDetail, reverse_definitions: _ReverseDefinitions
) -> IsqxExpr | None:
    if isinstance(meaning, QtyKind):
        return meaning.unit_si_coherent
    elif isinstance(meaning, (Dimensionless, Tagged)):
        return meaning
    elif isinstance(meaning, (LazyProduct, *_ARGS_NUMBER)):
        if (
            reverse_def := reverse_definitions.get(id(meaning))
        ) and reverse_def.annotated_metadata:
            return reverse_def.annotated_metadata.unit
        else:
            logger.warning(f"no unit found for {meaning} in registry")
            return None
    return None


def screaming_to_normal(s: str, i: int) -> str:
    # a better way would be to parse the docstring but that might be too verbose
    normalized = s.split(".")[-1].lower().replace("_", " ")
    return normalized.capitalize() if i == 0 else normalized


def inject_citation_into_docstring(
    obj: Object, agent: Visitor | Inspector
) -> None:
    if not obj.canonical_path.startswith("isqx._citations."):
        return
    # quick hack to put runtime value into the docstring
    doc_str = dynamic_import(obj.path)
    if not isinstance(doc_str, str):
        logger.warning(
            f"expected citation {obj.path} to be a string, got {type(doc_str)}"
        )
        return
    obj.docstring = Docstring(
        doc_str,
        parent=obj,
        parser=agent.docstring_parser,
        parser_options=agent.docstring_options,
    )



#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.9"
# dependencies = [
#     "jax",
#     "numpy",
#     "perfplot",
#     "numba",
#     "pint"
# ]
# ///
# mypy: ignore-errors
from pathlib import Path

import jax
import numpy as np
import perfplot
from airspeed import eas, eas_pure_py, ureg
from numba import njit


def setup(n):
    tas_np = np.random.rand(n) * 250 + 50
    density_np = np.random.rand(n) * 0.825 + 0.4
    return tas_np, density_np


def run_pint_quantity_in(tas_np, density_np):
    tas_pint = tas_np * (ureg.meter / ureg.second)
    density_pint = density_np * (ureg.kilogram / ureg.meter**3)
    result_pint = eas(tas_pint, density_pint)
    return result_pint.magnitude


def run_pint_numpy_in(tas_np, density_np):
    result_pint = eas(tas_np, density_np)
    return result_pint.magnitude  # NOTE: wrong units


def run_numpy(tas_np, density_np):
    return eas_pure_py(tas_np, density_np)


eas_jax_jit_cpu = jax.jit(eas_pure_py, backend="cpu")
eas_numba_jit = njit(eas_pure_py, parallel=True, nogil=True, cache=True)


def run_jax_cpu(tas_np, density_np):
    return eas_jax_jit_cpu(tas_np, density_np)


def run_numba(tas_np, density_np):
    return eas_numba_jit(tas_np, density_np)


def main(path_root: Path = Path(__file__).parent):
    n_range = [2**k for k in range(2, 27)]
    for n in n_range:  # precompile
        run_jax_cpu(*setup(n))
        run_numba(*setup(n))
    out = perfplot.bench(
        setup=setup,
        kernels=[
            run_pint_quantity_in,
            run_pint_numpy_in,
            run_numpy,
            run_jax_cpu,
            run_numba,
        ],
        n_range=n_range,
        labels=[
            "pint (Quantity inputs)",
            "pint",
            "baseline (np.ndarray inputs)",
            "baseline + jax.jit",
            "baseline + numba.njit",
        ],
        xlabel="np.ndarray size",
        equality_check=np.allclose,
    )
    out.save(path_root / "bench.png", transparent=True, bbox_inches="tight")
    out.show()


if __name__ == "__main__":
    main()
